from collections import defaultdict
from typing import Callable

import numpy as np

from ieml.ieml_objects.terms.tools import term
from ieml.script.tools import factorize


class Table:
    def __init__(self, cells, factorize_fn):
        super().__init__()

        if not isinstance(cells, np.ndarray) or cells.ndim != 2:
            raise ValueError("Must provide a 2d numpy array to create a Table object")

        if not isinstance(factorize_fn, Callable):
            raise ValueError("Must provide a function to factorize the cells into the headers")

        self.cells = cells
        self.factorize_fn = factorize_fn

        self.dim = 2 if all(s != 1 for s in self.cells.shape) else 1

        self.rows = None
        self.columns = None
        self.head = None
        self._build_headers()

    def _build_headers(self):
        if self.dim == 1:
            self.rows = []
            self.columns = []
            self.head = self.factorize_fn(list(self))
        else:
            self.rows = [self.factorize_fn(list(self.cells[i,:])) for i in range(self.shape[0])]
            self.columns = [self.factorize_fn(list(self.cells[:, j])) for j in range(self.shape[1])]
            self.head = self.factorize_fn(self.rows)

    @property
    def shape(self):
        return self.cells.shape

    def index(self, item):
        if item not in self:
            return []

        if self._index is None:
            self._index = {c: tuple(*zip(*np.where(self.cells == c))) for c in self}

        return [self._index[c] for c in item]

    def project(self, elements, key):
        result = defaultdict(list)

        for cell in self:
            result[cell] = [e for e in elements if key(cell, e)]

        return result

    def __eq__(self, other):
        return isinstance(other, Table) and self.head == other.head

    def __hash__(self):
        return self.head.__hash__()

    def __contains__(self, item):
        return item in self.head

    def __iter__(self):
        return np.nditer(self.cells, flags=['refs_ok'])


def get_tables_for_term(t):
    return [Table(cells[:,:,i], factorize_fn=factorize) for cells in term(t).script.cells
            for i in range(cells.shape[2])]

if __name__ == '__main__':
    t = get_tables_for_term('M:O:.M:M:.-')
    print([str(c) for c in t[0]])


from collections import namedtuple
import numpy as np

from ieml.commons import LANGUAGES
from ieml.ieml_objects.commons import IEMLObjects
from ieml.ieml_objects.terms.relations import Relations
from ieml.script.operator import script
from ieml.script.tools import factorize

Translations = namedtuple('Translations', list(LANGUAGES))
Translations.__getitem__ = lambda self, item: self.__getattribute__(item) if item in LANGUAGES \
    else tuple.__getitem__(self, item)


class Term(IEMLObjects):
    closable = True

    def __init__(self, s, dictionary):
        self.dictionary = dictionary
        self.script = script(s)

        self.grammatical_class = self.script.script_class

        super().__init__([])
        self.relations = Relations(term=self, dictionary=self.dictionary)
        self.index = None

        # if term in a dictionary, those values will be set
        self._translations = None
        self._root = None

        from ieml.ieml_objects.terms.tools import TermNotFoundInDictionary

    __hash__ = IEMLObjects.__hash__

    def __eq__(self, other):
        if not isinstance(other, Term):
            return False

        return self.script == other.script

    def _do_gt(self, other):
        return self.script > other.script

    def compute_str(self, children_str):
        return "[" + str(self.script) + "]"

    @property
    def parent(self):
        if self in self.dictionary.parents:
            return self.dictionary.parents[self]

        return None

    @property
    def partitions(self):
        return self.dictionary.partitions[self]

    @property
    def inhibitions(self):
        return self.dictionary.inhibitions[self.root]

    @property
    def translations(self):
        if self._translations is None:
            self._translations = Translations(**{l: self.dictionary.translations[l][self] for l in LANGUAGES})
        return self._translations

    @property
    def root(self):
        if self._root is None:
            self._root = self.dictionary.get_root(self.script)
        return self._root

    @property
    def rank(self):
        if self not in self.dictionary.ranks:
            print(self)
        return self.dictionary.ranks[self]

    @property
    def empty(self):
        return self.script.empty

    @property
    def defined(self):
        return all(self.__getattribute__(p) is not None for p in
                   ['translation', 'inhibitions', 'root', 'index', 'relations', 'rank'])

    @property
    def ntable(self):
        return sum(self.script.cells[i].shape[2] for i in range(len(self.script.cells)))

    def cells(self):
        if self.ntable != 1:
            raise ValueError("Too many dimension to generate table.")

        return self.script.cells[0][:,:,0]

    def headers(self):
        cells = self.cells()

        rows = [factorize([s for s in c]) for c in cells]
        columns = [factorize([s for s in c]) for c in cells.transpose()]

        return rows, columns

    @property
    def tables(self):
        return self.script.tables

    @property
    def singular_sequences(self):
        from .tools import term
        return [term(ss, dictionary=self.dictionary) for ss in self.script.singular_sequences]

    @property
    def layer(self):
        return self.script.layer

    def __contains__(self, item):
        from .tools import term
        if not isinstance(item, Term):
            item = term(item, dictionary=self.dictionary)
        elif item.dictionary != self.dictionary:
            print("\t[!] Comparison between different dictionary.")
            return False

        return item.script in self.script

    def __len__(self):
        return self.script.cardinal





class InvalidPathException(Exception):
    def __init__(self, element, path):
        self.element = element
        self.path = path

    def __str__(self):
        return "Can't access %s in %s, the path is invalid."%(str(self.path), str(self.element))


class CannotParse(Exception):
    def __init__(self, s):
        self.s = s

    def __str__(self):
        return "Unable to parse the following string %s."%str(self.s)


import itertools as it
from collections import defaultdict
from fractions import Fraction

from bidict import bidict
from ieml.ieml_objects import Term, Word, Sentence, SuperSentence, Morpheme, Text, Hypertext
from ieml.ieml_objects.hypertexts import Hyperlink
from ieml.usl import usl
from ieml.script.operator import sc
from ieml.script.tables import get_table_rank
from models.relations.relations import RelationsConnector
from models.relations.relations_queries import RelationsQueries

categories = bidict({Term: 1, Word: 2, Sentence: 3, SuperSentence: 4, Text: 5, Hypertext: 6})


def distance(uslA, uslB, weights):

    eo_total = sum(set_proximity_index(i, uslA, uslB) for i in categories if i != Hypertext) / (len(categories) - 1)
    oo_total = sum(object_proximity_index(i, uslA, uslB) for i in categories if i != Term) / (len(categories) - 1)

    return (eo_total + oo_total) / 2


def set_proximity_index(stage, uslA, uslB):
    """

    Parameters
    ----------
    stage
    uslA
    uslB

    Returns
    -------
    Proximity index between two sets of objects of the same layer
    """
    stages_A, children_A, children_multi_A = compute_stages(uslA)
    stages_B, children_B, children_multi_B = compute_stages(uslB)

    if len(stages_A[stage] | stages_B[stage]) == 0:
        return 1.0

    return float(len(stages_A[stage] & stages_B[stage])) / len(stages_A[stage] | stages_B[stage])


def object_proximity_index(stage, uslA, uslB):
    """

    Parameters
    ----------
    stage
    uslA
    uslB

    Returns
    -------
    Proximity index for 2 objects of the same layer
    """
    stages_A, children_A, children_multi_A = compute_stages(uslA)
    stages_B, children_B, children_multi_B = compute_stages(uslB)

    if stage is Term:
        raise ValueError

    size = float(len(stages_A[stage]) * len(stages_B[stage]))
    accum = 0.0
    for a, b in it.product(stages_A[stage], stages_B[stage]):
        accum += len(children_A[a] & children_B[b]) / (size * len(children_A[a] | children_B[b]))

    return accum


def connexity_index(stage, uslA, uslB):

    stages_A, children_A, children_multi_A = compute_stages(uslA)
    stages_B, children_B, children_multi_B = compute_stages(uslB)

    size = float(len(stages_A[stage]) * len(stages_B[stage]))
    accum = 0.0
    values = []

    for a, b in it.product(stages_A[stage], stages_B[stage]):

        if stage == Sentence or stage == SuperSentence:
            intersection = set(a.tree_graph.nodes) & set(b.tree_graph.nodes)
        else:
            intersection = children_A[a] & children_B[b]

        graph = build_graph(a, b, intersection)
        partitions = partition_graph(graph)
        values.append(connexity(partitions, intersection))

    return sum(values) / size


def mutual_inclusion_index(uslA, uslB):

    stages_A, children_A, children_multi_A = compute_stages(uslA)
    stages_B, children_B, children_multi_B = compute_stages(uslB)

    result = {Word: 0, Sentence: 0, SuperSentence: 0, Text: 0, Hypertext: 0}
    for a_st, b_st in it.permutations(categories.keys(), 2):
        size = float(len(stages_A[a_st]) * len(stages_B[b_st]))
        accum = 0.0

        # if true b in a
        direct = categories[a_st] > categories[b_st]

        for a, b in it.product(stages_A[a_st], stages_B[b_st]):
            if direct:
                accum += children_multi_A[a][b] / \
                         (size * len([e for e in children_multi_A[a] if e.__class__ == b.__class__]))
            else:
                accum += children_multi_B[b][a] / \
                         (size * len([e for e in children_multi_B[b] if e.__class__ == a.__class__]))

        if direct:
            result[a_st] += accum / 2.0
        else:
            result[b_st] += accum / 2.0
    return sum(result.values())/len(result)  # Return mean of the matrix cell index values


def compute_stages(usl):
    """
    Get all the elements in the usl by stages, Term, Word, Sentence, SSentence, Text, Hypertext
    :param usl:
    :return:
    """

    def usl_iter(usl):
        yield from usl.ieml_object.tree_iter()

    stages = {c: set() for c in categories}

    for e in (k for k in usl_iter(usl) if isinstance(k, tuple(categories))):
        stages[e.__class__].add(e)

    children = {}
    for cat in stages:
        for e in stages[cat]:
            if isinstance(e, Term):
                children[e] = set()
                continue

            if isinstance(e, (Word, Sentence, SuperSentence, Text, Hypertext)):
                _class = categories.inv[categories[e.__class__] - 1]
                children[e] = set(i for i in e.tree_iter() if isinstance(i, _class))

    result = defaultdict(lambda: defaultdict(lambda: 0))
    stack = []
    for e in usl_iter(usl):
        if e.__class__ not in categories:
            continue

        if stack:
            while categories[e.__class__] >= categories[stack[-1].__class__]:
                stack.pop()

        for k in stack:
            result[k][e] += 1

        stack.append(e)

    return stages, children, result


def flatten_dict(dico):

    lineage = []
    if isinstance(dico, list):
        return dico
    for child in dico:
        lineage.extend(flatten_dict(dico[child]))
    return lineage


def partition_graph(graph):

    partitions = []
    explored = set()

    def _explore_graph(node, graph, p):

        if node not in explored:
            p.add(node)
        explored.add(node)
        for adjacent_node in graph[node]:
            if adjacent_node not in explored:
                _explore_graph(adjacent_node, graph, p)
        return

    for node in graph:
        p = set()
        _explore_graph(node, graph, p)

        if p:
            partitions.append(p)

    return partitions


def build_graph(object_set_a, object_set_b, intersection):
    # See IEML pour l'ingenieur section 6.7.4.2 for the graph building algorithm

    graph = {node: [] for node in intersection}

    if isinstance(object_set_a, (Text, Hypertext)):

        combos = it.combinations(intersection, 2)
        for combination in combos:
            if combination[0].__class__ == combination[1].__class__:
                graph = _build_proposition_graph(combination, graph)
            elif combination[0].__class__ < combination[1].__class__:
                graph = _build_proposition_graph(combination, graph)
            elif combination[0].__class__ > combination[1].__class__:
                graph = _build_proposition_graph(combination, graph)

    if isinstance(object_set_a, (Sentence, SuperSentence)):
        # In this case the nodes are of type Word

        node_pairs = it.combinations(intersection, 2)

        for pair in node_pairs:
            node_1_addr_a = object_set_a.tree_graph.nodes.index(pair[0])
            node_2_addr_a = object_set_a.tree_graph.nodes.index(pair[1])
            node_1_addr_b = object_set_b.tree_graph.nodes.index(pair[0])
            node_2_addr_b = object_set_b.tree_graph.nodes.index(pair[1])

            if ((object_set_a.tree_graph.array[node_1_addr_a][node_2_addr_a] or
                 object_set_a.tree_graph.array[node_2_addr_a][node_1_addr_a]) and
                (object_set_b.tree_graph.array[node_1_addr_b][node_2_addr_b] or
                 object_set_b.tree_graph.array[node_2_addr_b][node_1_addr_b])):

                graph[pair[0]].append(pair[1])
                graph[pair[1]].append(pair[0])

    if isinstance(object_set_a, Word):
        # The nodes in the intersection set are Terms

        combos = it.combinations(intersection, 2)

        for combination in combos:
            if combination[0] in object_set_a.root.children and combination[1] in object_set_a.flexing.children and \
               combination[0] in object_set_b.root.children and combination[1] in object_set_b.flexing.children:
                graph[combination[0]].append(combination[1])
                graph[combination[1]].append(combination[0])
            elif combination[0] in object_set_a.flexing.children and combination[1] in object_set_a.root.children and \
                 combination[0] in object_set_b.flexing.children and combination[1] in object_set_b.root.children:
                graph[combination[0]].append(combination[1])
                graph[combination[1]].append(combination[0])

    return graph


def parents_index(usl_a, usl_b, index="EO"):

    parents_a, parents_b = get_parents(usl_a, usl_b)

    if index == 'EO':
        return len(parents_a & parents_b) / len(parents_a | parents_b)
    elif index == 'OO':
        return sum(len(a & b)/len(a | b) for a, b in it.product(parents_a, parents_b))/(len(parents_a) * len(parents_b))

    else:
        # TODO: create an invalid index exception and throw it here
        print("Wrong index")


def _build_proposition_graph(combination, graph):

    if isinstance(combination[0], (Sentence, SuperSentence)):
        prop_a = {elem for elem in combination[0].tree_iter() if isinstance(elem, Word)}
        prop_b = {elem for elem in combination[1].tree_iter() if isinstance(elem, Word)}
        if prop_a <= prop_b or prop_b <= prop_a or prop_a & prop_b:
            graph[combination[0]].append(combination[1])
            graph[combination[1]].append(combination[0])
    elif isinstance(combination[0], Word):
        prop_a = {elem for elem in combination[0].tree_iter() if isinstance(elem, Term)}
        prop_b = {elem for elem in combination[1].tree_iter() if isinstance(elem, Term)}
        if prop_a <= prop_b or prop_b <= prop_a or prop_a & prop_b:
            graph[combination[0]].append(combination[1])
            graph[combination[1]].append(combination[0])
    return graph


def get_parents(usl_a, usl_b):

    stages_A, children_A, children_multi_A = compute_stages(usl_a)
    stages_B, children_B, children_multi_B = compute_stages(usl_b)

    def tupleize(arr):
        return frozenset({(elem, arr.count(elem)) for elem in arr})

    rc = RelationsConnector()
    parents_a = {tupleize(flatten_dict(RelationsQueries.relations(term.script, 'FATHER_RELATION'))) for term in stages_A[Term]}
    parents_b = {tupleize(flatten_dict(RelationsQueries.relations(term.script, 'FATHER_RELATION'))) for term in stages_B[Term]}

    return parents_a, parents_b


def get_paradigms(object_set_a):

    if isinstance(object_set_a, Word) or isinstance(object_set_a, Sentence) or isinstance(object_set_a, SuperSentence):
        term_list = [elem for elem in object_set_a.tree_iter() if isinstance(elem, Term)]
    else:
        stages_A, children_A, children_multi_A = compute_stages(object_set_a)
        term_list = stages_A[Term]

    paradigms = {i: [] for i in range(1, 6)}

    rc = RelationsConnector()

    for term in term_list:
        for paradigm in RelationsQueries.relations(term.script, "CONTAINED"):
            paradigms[RelationsQueries.rank(paradigm)].append(paradigm)

    return paradigms


def get_grammar_class(uslA, uslB):
    stages_A, children_A, children_multi_A = compute_stages(uslA)
    stages_B, children_B, children_multi_B = compute_stages(uslB)

    grammar_classes_A = [term.script.script_class for term in stages_A[Term]]
    grammar_classes_B = [term.script.script_class for term in stages_B[Term]]

    return grammar_classes_A, grammar_classes_B


def connexity(partitions, node_intersection):

    if len(node_intersection) == 0:
        return 0

    return Fraction(sum(len(p) for p in partitions if len(p) > 1), (len(partitions) * len(node_intersection)))


def print_graph(graph):

    for k, v in graph.items():
        print(str(k) + ": ")
        print("[", end="")
        for elem in v:
            print(str(elem), end=", ")
        print("]")


def list_intersection_cardinal(list_a, list_b):
    """
    The purpose of this method is to perform the intersection of two lists.
    The reason why we need this method is because we have a use case where we need to keed repeted element in memory
    and also use the list as a set.
    :return:
    """
    intersection_cardinal = 0

    if len(list_a) >= len(list_b):
        tmp = list_a
        iteration_list = list_b

    else:
        tmp = list_b
        iteration_list = list_a


    for element in iteration_list:
        if element in tmp:
            intersection_cardinal += 1
            tmp.remove(element)

    return intersection_cardinal


def list_union_cardinal(list_a, list_b):
    union_cardinal = len(list_a) + len(list_b)
    return union_cardinal


def grammatical_class_index(uslA, uslB, index, stage):
    """

    :param uslA:
    :param uslB:
    :param index:must be the set proximity index : "EO" or the object proximity index : "OO"
    :param stage:must be Term, Word, Sentence or SuperSentence
    :return:the value of the index
    """
    stages_A, children_A, children_multi_A = compute_stages(uslA)
    stages_B, children_B, children_multi_B = compute_stages(uslB)

    grammar_classes_a = [elem.grammatical_class for elem in stages_A[stage]]
    grammar_classes_b = [elem.grammatical_class for elem in stages_B[stage]]

    if index == 'EO':
        index_value = (list_intersection_cardinal(grammar_classes_a, grammar_classes_b) /
                      list_union_cardinal(grammar_classes_a, grammar_classes_b))

    elif index == 'OO':
        if stage is Term:
            raise ValueError

        size = float(len(stages_A[stage]) * len(stages_B[stage]))
        index_value = 0.0
        for a, b in it.product(stages_A[stage], stages_B[stage]):
            # we replace every child of A and B by their grammatical class, and create a list of it
            grammar_list_a = [e.grammatical_class for e in children_A[a]]
            grammar_list_b = [e.grammatical_class for e in children_B[b]]

            # we sort the list in order to do the intersection and union of lists
            grammar_list_a.sort()
            grammar_list_b.sort()

            index_value += (list_intersection_cardinal(grammar_list_a, grammar_list_b) /
                           (size * list_union_cardinal(grammar_list_a, grammar_list_b)))
    else:
        raise ValueError

    return index_value


#TODO complete this function with loulou functions
def paradigmatic_equivalence_class_index(uslA, uslB, paradigm_rank, index):
    """

    :param uslA:
    :param uslB:
    :param paradigm_rank: rank of the paradigm wanted, must be 1 (root paradigm), 2, 3, 4 or 5
    :param index: must be the set proximity index (EO) or the object proximity index (OO)
    :return the paradigmatic equivalence class of the choosen index (EO or OO)
    """
    index_value = 0

    stages_A, children_A, children_multi_A = compute_stages(uslA)
    stages_B, children_B, children_multi_B = compute_stages(uslB)
    paradigms_a = get_paradigms(uslA)
    paradigms_b = get_paradigms(uslB)

    #faire l inverse, d abord if table_rank puis if index ?
    if index == 'EO':
        index_value = (list_intersection_cardinal(paradigms_a[paradigm_rank], paradigms_b[paradigm_rank]) /
                       list_union_cardinal(paradigms_a[paradigm_rank], paradigms_b[paradigm_rank]))

    elif index == 'OO':
        #  We can only find paradigms of terms.
        # So we have to build the '00' matrix with the words of uslA and uslB as rows and columns
        size = float(len(stages_A[Word]) * len(stages_B[Word]))
        if size == 0:
            raise ValueError

        for a, b in it.product(stages_A[Word], stages_B[Word]):
            paradigms_a = get_paradigms(a)
            paradigms_b = get_paradigms(b)

            index_value += (list_intersection_cardinal(paradigms_a[paradigm_rank], paradigms_b[paradigm_rank]) /
                            list_union_cardinal(paradigms_a[paradigm_rank], paradigms_b[paradigm_rank]))
        index_value /= size

    return index_value


if __name__ == '__main__':
    a = "{/[([a.i.-]+[i.i.-])*([E:A:T:.]+[E:S:.wa.-]+[E:S:.o.-])]//[([([a.i.-]+[i.i.-])*([E:A:T:.]+[E:S:.wa.-]+[E:S:.o.-])]{/[([a.i.-]+[i.i.-])*([E:A:T:.]+[E:S:.wa.-]+[E:S:.o.-])]/}*[([t.i.-s.i.-'i.B:.-U:.-'we.-',])*([E:O:.wa.-])]*[([E:E:T:.])])+([([a.i.-]+[i.i.-])*([E:A:T:.]+[E:S:.wa.-]+[E:S:.o.-])]*[([t.i.-s.i.-'u.B:.-A:.-'wo.-',])]*[([E:T:.f.-])])]/}"
    b = usl(a)
    c = usl(a)
    print(distance(b, c, None))

    word_a = Word(Morpheme(
        [Term(sc('wa.')), Term(sc("l.-x.-s.y.-'")), Term(sc("e.-u.-we.h.-'")), Term(sc("M:.E:A:M:.-")),
         Term(sc("E:A:.k.-"))]), Morpheme([Term(sc('wo.')), Term(sc("T:.E:A:T:.-")), Term(sc("E:A:.k.-")),
                                           Term(sc("T:.-',S:.-',S:.-'B:.-'n.-S:.U:.-',_"))]))
    word_b = Word(Morpheme([Term(sc("l.-x.-s.y.-'")), Term(sc("e.-u.-we.h.-'"))]), Morpheme([Term(sc("T:.E:A:T:.-"))]))

    ht_a = Hypertext(Text([word_a]))
    ht_b = Hypertext(Text([word_b]))
    connexity_index(Word, ht_a, ht_b)

    # rc = RelationsConnector()
    # parser = rc.get_script("E:M:.M:O:.-")
    # d = flatten_dict(parser['RELATIONS']['FATHER_RELATION'])
    # print(d)
    #
    # term1 = Term(sc("T:.E:A:T:.-"))
    # term2 = Term(sc("e. - u. - we.h. - '"))
    # term3 = Term(sc("l. - x. - s.y. - '"))
    #
    # term1.check()
    # term2.check()
    # term3.check()
    #
    # graph = {term1: [term2, term3], term2: [term1], term3: [term2]}
    #
    # partition_graph(graph)



from collections import OrderedDict, defaultdict

NB_RELATIONS = 12

RELATIONS = ['contains',         # 0
             'contained',        # 1
             'father_substance', # 2
             'child_substance',  # 3
             'father_attribute', # 4
             'child_attribute',  # 5
             'father_mode',      # 6
             'child_mode',       # 7
             'opposed',          # 8
             'associated',       # 9
             'crossed',          # 10
             'twin',             # 11

             'table_0',
             'table_1',
             'table_2',
             'table_3',
             'table_4',
             'table_5',

             'inclusion',        # 12
             'father',           # 13
             'child',            # 14
             'etymology',        # 15
             'siblings',         # 16
             'table'             # 17
             ]

INVERSE_RELATIONS = {
    'father_substance': 'child_substance',
    'child_substance': 'father_substance',  # 3
    'father_attribute': 'child_attribute', # 4
    'child_attribute': 'father_attribute',  # 5
    'father_mode': 'child_mode',      # 6
    'child_mode': 'father_mode',
    'contains': 'contained',
    'contained': 'contains',
    'opposed':'opposed',          # 8
    'associated':'associated',       # 9
    'crossed': 'crossed',        # 10
    'twin': 'twin',
    'table_0': 'table_0',
    'table_1': 'table_1',
    'table_2': 'table_2',
    'table_3': 'table_3',
    'table_4': 'table_4',
    'table_5': 'table_5',
    'father': 'child',
    'child': 'father',
    'inclusion': 'inclusion',
    'etymology': 'etymology',        # 15
    'siblings': 'siblings',         # 16
    'table': 'table'
}


class Relations:
    def __init__(self, term, dictionary):
        super().__init__()

        self.dictionary = dictionary
        self.term = term
        self._all = None
        for reltype in RELATIONS:
            setattr(self, "_%s" % reltype, None)

    def all(self, dict=False):
        if self._all is None:
            rels = defaultdict(list)

            for reltype in RELATIONS:
                for t in self[reltype]:
                    rels[t].append(reltype)

            self._all = OrderedDict()

            for t in sorted(rels):
                self._all[t] = rels[t]

            self._all_tuple = tuple(self._all)

        if dict:
            return self._all
        else:
            return self._all_tuple

    def to(self, term, relations_types=None):
        if relations_types is None:
            relations_types = RELATIONS

        if term not in self.all(dict=True):
            return []

        relations = self.all(dict=True)[term]
        result = []
        for reltype in relations:
            if reltype in relations_types:
                result.append(reltype)

        return result

    def __iter__(self):
        return self.all().__iter__()

    def __len__(self):
        return len(self.all())

    def __contains__(self, item):
        return item in self.all(dict=True)

    def __getitem__(self, item):
        if isinstance(item, int):
            item = RELATIONS[item]

        if isinstance(item, str):
            return self.__getattribute__(item)

        raise NotImplemented


def get_relation(reltype):
    def getter(self):
        if getattr(self, "_%s" % reltype) is None:
            relations = tuple(self.dictionary.rel(reltype, term=self.term))
            setattr(self, "_%s" % reltype, relations)

        return getattr(self, "_%s" % reltype)

    return getter

for reltype in RELATIONS:
    setattr(Relations, reltype, property(fget=get_relation(reltype)))


if __name__ == '__main__':
    from ieml.ieml_objects.tools import term
    t = term("wa.")
    print([str(tt) for tt in t.relations.etymology])


from ieml.ieml_objects.terms.version import DictionaryVersion
from ieml.script import Script
from ieml.ieml_objects.terms.terms import Term
from ieml.ieml_objects.terms.dictionary import Dictionary


class TermNotFoundInDictionary(Exception):
    def __init__(self, term, dictionary):
        self.message = "Cannot find term %s in the dictionary %s" % (str(term), str(dictionary.version))

    def __str__(self):
        return self.message


def term(arg, dictionary=None):
    if isinstance(arg, Term):
        return arg

    if not isinstance(dictionary, Dictionary):
        if isinstance(dictionary, (str, DictionaryVersion)):
            dictionary = Dictionary(dictionary)
        else:
            dictionary = Dictionary()

    if isinstance(arg, int):
        return dictionary.index[arg]

    if isinstance(arg, str):
        if arg[0] == '[' and arg[-1] == ']':
            arg = arg[1:-1]

    if isinstance(arg, Script) or isinstance(arg, str):
        if arg in dictionary:
            return dictionary.terms[arg]

    raise TermNotFoundInDictionary(arg, dictionary)




import os
DICTIONARY_BUCKET_URL = "https://s3.amazonaws.com/ieml-dictionary-versions/"
DICTIONARY_VERSIONS_FOLDER = "data/dictionary/"

DICTIONARY_DEFAULT_VERSION = '2017-06-07_00:00:00'
DICTIONARY_FOLDER = os.path.join(os.path.dirname(__file__), "data/dictionary")

AWS_ACCESS_KEY_ID = ''
AWS_SECRET_ACCESS_KEY = ''



import collections
from collections import defaultdict
from itertools import chain

import numpy as np

from ieml.ieml_objects.terms import Term, term, TermNotFoundInDictionary
from ieml.script.operator import script
from ieml.script.script import Script

Cell = collections.namedtuple('Cell', ['row', 'column', 'coordinate', 'table', 'script', 'term'])

_terms = defaultdict(list)


class Table:
    def __init__(self, paradigm):
        super().__init__()

        self.paradigm = term(paradigm)
        if self.paradigm.script.cardinal == 1:
            raise ValueError("Table are defined on paradigm, not singular sequences.")

        self._index = {}

        cells = self.paradigm.cells()
        self.dim = 2 if all(s != 1 for s in cells.shape) else 1

        def _cell(s, i, j):
            try:
                t = term(s, dictionary=self.paradigm.dictionary)
            except TermNotFoundInDictionary:
                t = None

            row = None
            column = None

            if i is not None and j is not None:
                # singular sequence -> table cell
                if self.dim == 2:
                    row = self.rows[i]
                    column = self.columns[j]
                else:
                    column = self.header

            res = Cell(row=row,
                       column=column,
                       coordinate=(i, j),
                       table=self,
                       script=s,
                       term=t)

            if t is not None:
                _terms[t].append(res)

            self._index[s] = res
            return res

        self.header = _cell(self.paradigm.script, None, None)

        if self.dim == 2:
            rows, columns = self.paradigm.headers()
            self.rows = [_cell(s, i, None) for i, s in enumerate(rows)]
            self.columns = [_cell(s, None, j) for j, s in enumerate(columns)]

        self.cells = np.empty(shape=cells.shape, dtype=Cell)
        for i, column in enumerate(cells):
            for j, s in enumerate(column):
                self.cells[i, j] = _cell(s, i, j)

    @property
    def shape(self):
        return self.cells.shape

    def index(self, item):
        return self[item].coordinate

    def project(self, elements, key):
        result = {}
        for e in elements:
            result[e] = [c for c in self if key(c.element, e)]

        return result

    def __eq__(self, other):
        return isinstance(other, Table) and self.paradigm == other.paradigm

    def __hash__(self):
        return self.paradigm.__hash__()

    def __contains__(self, item):
        if isinstance(item, Term):
            return item in self.paradigm

        ss = None
        if isinstance(item, Script):
            ss = set(item.singular_sequences)

        # script list set
        if isinstance(item, collections.Iterable) and all(isinstance(ss, Script) for ss in item):
            ss = set(chain.from_iterable(s.singular_sequences for s in item))

        if ss is not None:
            return ss.issubset(self.paradigm.script.singular_sequences)

        raise NotImplemented

    def __iter__(self):
        return map(lambda e: e[()],
                   np.nditer(self.cells, flags=['refs_ok'], op_flags=['readonly']))

    def __getitem__(self, item):
        if isinstance(item, (Script, Term, str)):
            s = script(item)
            return self._index[s]

        if not isinstance(item, collections.Iterable) and self.dim == 1:
            if isinstance(item, int):
                return self.cells[item, 0]
            elif item is None:
                return self.header

        if item[0] is None:
            if item[1] is None:
                return self.header

            if self.dim != 2:
                raise KeyError("No columns in 1d table")
            return self.columns[item[1]]

        elif item[1] is None:
            if self.dim != 2:
                raise KeyError("No rows in 1d table")

            return self.rows[item[0]]
        else:
            return self.cells[item]

    def __str__(self):
        return "<Table %s, (%d, %d)>"%(str(self.paradigm), self.shape[0], self.shape[1])

    def __len__(self):
        return self.shape[0] * self.shape[1]


class RootTable:
    def __init__(self, root_term):

        if not isinstance(root_term, Term) or root_term.root != root_term:
            raise ValueError("Invalid object for root table creation, expected a root term, not %s."%str(root_term))

        self.term = root_term
        # self.tables =

if __name__ == '__main__':
    from ieml.ieml_objects.terms import term
    t = Table(term('M:O:.M:M:.-'))
    print(t[t.index(term("j.t.-"))].script)
    print(t)



import random

import itertools

import functools
from ieml.ieml_objects.commons import IEMLObjects
from ieml.ieml_objects.exceptions import InvalidIEMLObjectArgument
from ieml.ieml_objects.parser.parser import IEMLParser

from ieml.ieml_objects.sentences import Sentence, Clause, SuperSentence, SuperClause
from ieml.ieml_objects.texts import Text
from ieml.ieml_objects.words import Word, Morpheme
from ieml.ieml_objects.exceptions import CantGenerateElement
from ieml.ieml_objects.terms import Term, Dictionary


def ieml(arg):
    if isinstance(arg, IEMLObjects):
        return arg

    if isinstance(arg, str):
        return IEMLParser().parse(arg)

    raise ValueError("Invalid argument, c'ant instantiate an IEMLObject from %s."%str(arg))


def _loop_result(max_try):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            ex = None
            for i in range(max_try):
                try:
                    return func(*args, **kwargs)
                except InvalidIEMLObjectArgument as e:
                    ex = e
                    continue

            raise CantGenerateElement(str(ex))
        return wrapper
    return decorator


class RandomPoolIEMLObjectGenerator:
    def __init__(self, level=Text, pool_size=20):
        self.level = level
        self.pool_size = pool_size

        if level > Text:
            raise ValueError('Cannot generate object higher than a Text.')

        self._build_pools()

        self.type_to_method = {
            Term: self.term,
            Word: self.word,
            Sentence: self.sentence,
            SuperSentence: self.super_sentence,
            Text: self.text
        }

    def _build_pools(self):
        """
        Slow method, retrieve all the terms from the database.
        :return:
        """
        if self.level >= Word:
            # words
            self.words_pool = set(self.word() for i in range(self.pool_size))

        if self.level >= Sentence:
            # sentences
            self.sentences_pool = set(self.sentence() for i in range(self.pool_size))

        if self.level >= SuperSentence:
            self.super_sentences_pool = set(self.super_sentence() for i in range(self.pool_size))

        if self.level >= Text:
            self.propositions_pool = set(itertools.chain.from_iterable((self.words_pool, self.sentences_pool, self.super_sentences_pool)))

        # self.hypertext_pool = set(self.hypertext() for i in range(self.pool_size))

    @_loop_result(10)
    def term(self):
        return random.sample(Dictionary().index, 1)[0]

    @_loop_result(10)
    def uniterm_word(self):
        return Word(Morpheme(random.sample(Dictionary().index, 1)))

    @_loop_result(10)
    def word(self):
        return Word(Morpheme(random.sample(Dictionary().index, 3)), Morpheme(random.sample(Dictionary().index, 2)))

    def _build_graph_object(self, primitive, mode, object, max_nodes=6):
        nodes = {primitive()}
        modes = set()

        if max_nodes < 2:
            raise ValueError('Max nodes >= 2.')

        result = set()

        for i in range(random.randint(2, max_nodes)):
            while True:
                s, a, m = random.sample(nodes, 1)[0], primitive(), mode()
                if a in nodes or m in nodes or a in modes:
                    continue

                nodes.add(a)
                modes.add(m)

                result.add(object(s, a, m))
                break
        return result

    @_loop_result(10)
    def sentence(self, max_clause=6):
        def p():
            return random.sample(self.words_pool, 1)[0]

        return Sentence(self._build_graph_object(p, p, Clause, max_nodes=max_clause))

    @_loop_result(10)
    def super_sentence(self, max_clause=4):
        def p():
            return random.sample(self.sentences_pool, 1)[0]

        return SuperSentence(self._build_graph_object(p, p, SuperClause, max_nodes=max_clause))

    @_loop_result(10)
    def text(self):
        return Text(random.sample(self.propositions_pool, random.randint(1, 8)))

    def from_type(self, type):
        try:
            return self.type_to_method[type]()
        except KeyError:
            raise ValueError("Can't generate that type or not an IEMLObject : %s"%str(type))


import bidict


class TreeStructure:
    def __init__(self):
        super().__init__()
        self._str = None
        self._paths = None
        self.children = None  # will be an iterable (list or tuple)

    def __str__(self):
        return self._str

    def __ne__(self, other):
        return not self.__eq__(other)

    def __eq__(self, other):
        if not isinstance(other, (TreeStructure, str)):
            return False

        return self._str == str(other)

    def __hash__(self):
        """Since the IEML string for any proposition AST is supposed to be unique, it can be used as a hash"""
        return self.__str__().__hash__()

    def __iter__(self):
        """Enables the syntactic sugar of iterating directly on an element without accessing "children" """
        return self.children.__iter__()

    def tree_iter(self):
        yield self
        for c in self.children:
            yield from c.tree_iter()

LAYER_MARKS = [
    ':',
    '.',
    '-',
    '\'',
    ',',
    '_',
    ';'
]

GRAMMATICAL_CLASS_NAMES = bidict.bidict({
    0: 'AUXILIARY',
    1: 'VERB',
    2: 'NOUN'
})

LANGUAGES = {
    'fr',
    'en'
}


from ieml.ieml_objects.commons import IEMLObjects
from ieml.ieml_objects.exceptions import InvalidIEMLObjectArgument
from ieml.ieml_objects.sentences import Sentence, SuperSentence
from ieml.ieml_objects.words import Word


class Text(IEMLObjects):
    closable = True

    def __init__(self, children):
        try:
            _children = [e for e in children]
        except TypeError:
            raise InvalidIEMLObjectArgument(Text, "The argument %s is not iterable." % str(children))

        if not all(isinstance(e, (Word, Sentence, SuperSentence)) for e in _children):
            raise InvalidIEMLObjectArgument(Text, "Invalid type instance in the list of a text,"
                                                  " must be Word, Sentence or SuperSentence.")

        super().__init__(sorted(set(_children)))

    def compute_str(self, children_str):
        return '{/' + '//'.join(children_str) + '/}'



from collections import namedtuple, OrderedDict

import itertools
import numpy as np
from ieml.script.operator import script

Tab = namedtuple('Tab', ['rows', 'columns', 'paradigm', 'cells'])


class Table:
    def __init__(self, cells):
        if not isinstance(cells, np.ndarray):
            raise ValueError("The cells argument must be an array of singular "
                             "sequences, not %s"%cells.__class__.__name__)

        if len(cells.shape) != 3:
            raise ValueError("Invalid shape for cells")

        self.cells = cells

        self.headers = None
        self.tabs = None
        self.build_headers()

        self.paradigm = script(self.headers)
        self._index = None
        self._dim = None

        self.rank = None
        self.term = None
        self.partition = None

    def define_table(self, term, table=None):
        if not term.defined:
            raise ValueError("Term %s is not defined in the dictionary, can't define a table with it."%str(term))

        self.term = term
        self.rank = term.rank

        if self.rank != 1 and not table.defined:
            raise ValueError("Table is not defined in the dictionary, can't define a table with it as partition."%str(term))

        self.partition = table

    @property
    def defined(self):
        return all(self.__getattribute__(p) is not None for p in ['rank', 'term'])

    def build_headers(self):
        self.headers = {}
        self.tabs = []

        for t in self.cells.transpose(2,0,1):
            rows = [script(c) for c in t]
            columns = [script(c) for c in t.transpose()]
            tabs_sc = script(rows)
            tab = Tab(rows=rows, columns=columns, paradigm=tabs_sc, cells=t)
            self.headers[tabs_sc] = tab
            self.tabs.append(tab)

    def index(self, s):
        if s not in self.paradigm:
            return []

        if self._index is None:
            self._index = {ss: tuple(*zip(*np.where(self.cells == ss))) for ss in self.paradigm.singular_sequences}

        return [self._index[ss] for ss in s.singular_sequences]

    @property
    def dim(self):
        if self._dim is None:
            self._dim = 3
            if self.cells.shape[2] == 1:
                self._dim -= 1

            if self.cells.shape[1] == 1:
                self._dim -= 1

            if self.cells.shape[0] == 1:
                self._dim -= 1

        return self._dim

    def all(self):
        """

        :return: all script referenced in this table
        """
        iter_list = []
        for tab in self.tabs:
            iter_list.append(tab.rows)
            iter_list.append(tab.columns)
            iter_list.append([tab.paradigm])
        iter_list.append(self.cells.flatten())
        iter_list.append([self.paradigm])

        return set(itertools.chain.from_iterable(iter_list))

    def __eq__(self, other):
        return isinstance(other, Table) and self.paradigm == other.paradigm

    def __hash__(self):
        return self.paradigm.__hash__()




from .words import Morpheme, Word
from .sentences import Clause, Sentence, SuperClause, SuperSentence
from .texts import Text
from .terms import Term
from .hypertexts import Hyperlink, Hypertext, PropositionPath
from .tools import RandomPoolIEMLObjectGenerator, ieml


import os
from itertools import groupby

from ieml.ieml_objects.terms.version import get_available_dictionary_version

from ieml.ieml_objects.terms import Dictionary, Table

from scipy.sparse.csgraph._shortest_path import shortest_path

from ieml.ieml_objects.terms.tools import term, TermNotFoundInDictionary

import numpy as np

from ieml.script.constants import MAX_LAYER
from ieml.script.script import MultiplicativeScript, NullScript, AdditiveScript

__distance_mat = {}


def get_distance_matrix(version):
    if version not in __distance_mat:
        __distance_mat[version] = load_distance_matrix(version)

    return __distance_mat[version]


def load_distance_matrix(version):
    FILE = '/tmp/distance_m_%s.npy'%str(version)
    FILE_w = '/tmp/distance_m_w_%s.npy'%str(version)

    if os.path.isfile(FILE) and os.path.isfile(FILE_w):
        return np.load(FILE), np.load(FILE_w)
    else:
        graph_ethy = Dictionary(version).relations_graph({
            'etymology': 1.0, # 1 to 0 (1/(layer0 - layer1)**2
            'inclusion': 1.0, # 0 or 1
            'siblings' : 1.5, # 0 or 1
            'table'    : 1/3  # 0 to 6
        })
        # graph_ethy_m = Dictionary().relations_graph(['etymology', 'inclusion', 'siblings',
        #                                            'table_0', 'table_1', 'table_2', 'table_3',
        #                                            'table_4', 'table_5']).astype(np.float32).todense()

        graph_ethy = 1.0/graph_ethy

        graph_ethy[graph_ethy == np.inf] = 0

        # dist_m = shortest_path(graph_ethy, directed=False, unweighted=True)
        dist_m_w, pred = shortest_path(graph_ethy, directed=False, unweighted=False, return_predecessors=True)
        np.save(FILE, pred)
        np.save(FILE_w, dist_m_w)
        # dist_m = None
        return pred, dist_m_w


def _distance_etymology(term0, term1):
    i = 0.0
    index = term1.index
    pred = get_distance_matrix(term0.dictionary.version)[0]
    while index != term0.index:
        i += 1.0
        index = pred[term0.index, index]
        if index == -9999:
            return 9999

    return i


def _nb_relations(term0, term1):
    return get_distance_matrix(term0.dictionary.version)[1][term0.index, term1.index]


def _max_rank(term0, term1):
    rel_table = term0.relations.to(term1, ['table_%d'%i for i in range(1, 6)])
    # print(rel_table)
    if 'table_5' in rel_table:
        return 0

    if 'table_4' in rel_table:
        return 0

    if 'table_3' in rel_table:
        return 1

    if 'table_2' in rel_table:
        return 2

    if 'table_1' in rel_table:
        return 3

    return 4
    # return 4 + abs(term0.script.layer - term1.script.layer)


def distance(term0, term1):
    if term0 == term1:
        return 0.,0.,0.

    return _distance_etymology(term0, term1), _nb_relations(term0, term1), _max_rank(term0, term1)



# def _test_diagram(t):
#     print("Diagram for term %s -- %s"%(str(t), t.translations.fr))
#
#     other = sorted((distance(t, t1) ,t1) for t1 in Dictionary() if t1 != t)
#
#     cat = defaultdict(list)
#     for d, tt in other:
#         for rel in t.relations.to(tt, relations_types=['father', 'child', 'contains', 'contained', 'table', 'siblings']):
#             cat[rel].append((d, tt))
#
#     for rel in cat:
#         cat[rel] = [(d,tt) for d, tt in sorted(cat[rel]) if d[1] <= 1.0]
#
#     for k, v in cat.items():
#         print("\t[%s]"%k)
#         for d, tt in v:
#             print("%s (%.2f, %.2f, %.2f) - %s [%s]" % (str(tt), d[0], d[1], d[2], tt.translations.fr,
#                                                        ', '.join(t.relations.to(tt))))


def _test_term(t, distance_f=distance):
    print("Distance from term %s -- %s"%(str(t), t.translations['fr']))

    other = sorted((distance_f(t, t1) ,t1) for t1 in Dictionary() if not t1.script.paradigm)
    kkk = [t for t in other if t[0][1] < 2.0]
    for d, tt in kkk[:30]:
        print("%s (%.2f, %.2f, %.2f) - %s [%s]"%(str(tt), d[0], d[1], d[2], tt.translations['fr'],
                                                 ', '.join(t.relations.to(tt))))

__graph_ethy = None
def distance2(t0, t1):
    global __graph_ethy
    if __graph_ethy is None:
        __graph_ethy = Dictionary().relations_graph({
            'etymology': 1.0, # 1 to 0 (1/(layer0 - layer1)**2
            'inclusion': 1.0, # 0 or 1
            'siblings' : 1.5, # 0 or 1
            'table'    : 1/3  # 0 to 6
        })
        __graph_ethy = np.exp(-__graph_ethy)
        for i in range(__graph_ethy.shape[0]):
            __graph_ethy[i, i] = 0

    v = __graph_ethy[t0.index, t1.index] - __graph_ethy[t1.index, :]
    return __graph_ethy[t0.index, t1.index]#np.sum(np.dot(v, v.transpose()))


def _test_term2(t):
    print("Distance from term %s -- %s, with tanh distance"%(str(t), t.translations['fr']))

    other = sorted((distance2(t, t1), t1) for t1 in Dictionary() if not t1.script.paradigm)
    kkk = [t for t in other if t[0] < 1.0]
    for d, tt in kkk[:30]:
        print("%s (%.2f) - %s [%s]"%(str(tt), d, tt.translations['fr'], ', '.join(t.relations.to(tt))))

mat_distance3 = None

def _table_entry(t):
    table = Table(t)
    return {
        'table': table,
        'term': t,
        'ss_set': set(t.script.singular_sequences)
    }



def _order(t0, t1):
    return (t0, t1) if t0 < t1 else (t1, t0)

C = .9

_siblings = {
    'opposed': 2, 'crossed': 1, 'associated': 3, 'twin': 0
}

def _get_sibling(t0, t1):
    return max(map(lambda r: _siblings[r], t0.relations.to(t1, ['opposed', 'crossed', 'associated', 'twin']))) + .5

# layer -> rank -> terms[]
layers_by_rank = None


def _get_greater_common_rank(t0, t1):
    global layers_by_rank
    if layers_by_rank is None:
        layers_by_rank = {
            i: {k: [_table_entry(t) for t in g]
                for k, g in groupby(sorted([t for t in layer if t.ntable == 1 and len(t) != 1], key=lambda t: t.rank),
                                    key=lambda t: t.rank)}
            for i, layer in enumerate(t0.dictionary.layers)
        }

    ranks = layers_by_rank[t0.layer]
    max_rank = max(ranks)

    ss_set = set(t0.script.singular_sequences).union(t1.script.singular_sequences)

    for i in range(max_rank, -1, -1):
        if i not in ranks:
            continue

        if any(ss_set.issubset(tt['ss_set']) for tt in ranks[i]):
            return float(max(i, 1)) / max_rank

    return 1.0/5.0


def _distance_same_layer(t0, t1):
    t0, t1 = _order(t0, t1)

    if t0.layer != t1.layer:
        raise ValueError("Nop nop different layer")

    if t0 == t1:
        return 1.0

    if t0.root == t1.root:
        # siblings
        if t1 in t0.relations.siblings:
            return C + (1.0 - C) * float(_get_sibling(t0, t1)) / 4.0

        # table

        return C * _get_greater_common_rank(t0, t1)

    return 0.0


def _distance_upper_layer(t0, t1):
    t0, t1 = _order(t0, t1)

    if isinstance(t1.script, MultiplicativeScript):
        a, b, c = t1.script.children

        t_a = term(a)

        if isinstance(c, NullScript):
            if isinstance(b, NullScript):
                return mat_distance3[t0.index, t_a.index]
            else:
                t_b = term(b)

                return 0.6 * mat_distance3[t0.index, t_a.index] + \
                       0.4 * mat_distance3[t0.index, t_b.index]
        else:
            t_b = term(b)
            t_c = term(c)

            return 0.5 * mat_distance3[t0.index, t_a.index] + \
                   0.3 * mat_distance3[t0.index, t_b.index] + \
                   0.2 * mat_distance3[t0.index, t_c.index]
    elif isinstance(t1.script, AdditiveScript):
        res = [_distance_upper_layer(t0, term(s)) for s in t1.script]

        return sum(res) / len(res)


def _distance_upper_layer2(t0, t1):
    if isinstance(t1.script, MultiplicativeScript):
        a, b, c = t1.script.children

        def _dist(s):
            if isinstance(s, NullScript):
                return 0.0

            try:
                return mat_distance3[t0.index, term(s, dictionary=t0.dictionary).index]
            except TermNotFoundInDictionary:
                return 0.0

        d_a = _dist(a)
        d_b = _dist(b)
        d_c = _dist(c)

        if isinstance(c, NullScript):
            if isinstance(b, NullScript):
                return 0.80 * d_a
            else:
                return 0.40 * d_a + 0.30 * d_b
        else:
            return 0.35 * d_a + 0.20 * d_b + 0.10 * d_c

    elif isinstance(t1.script, AdditiveScript):
        res = [_distance_upper_layer2(t0, term(s, dictionary=t0.dictionary)) for s in t1.script]

        return sum(res) / len(res)


def _compute_distance_3(dictionary):
    global mat_distance3

    # indexes_layers = {
    #     k: [t.index for t in g] for k, g in groupby(dictionary, key=lambda t: t.layer)
    # }

    def _set_distance(t0, t1, dist):
        mat_distance3[t0.index, t1.index] = dist
        mat_distance3[t1.index, t0.index] = dist

    mat_distance3 = np.zeros(shape=[len(dictionary)]*2, dtype=float)

    for root in dictionary.roots:
        terms = root.relations.contains
        for i, t0 in enumerate(terms):
            for t1 in terms[i:]:
                _set_distance(t0, t1, _distance_same_layer(t0, t1))

    for j in range(1, MAX_LAYER):
        for i, layer0 in enumerate(dictionary.layers[:MAX_LAYER - 1]):
            if i+j == len(dictionary.layers):
                break

            for t0 in layer0:
                for t1 in (tt for tt in dictionary.rel('child', t0) if tt.layer == i + j):
                    _set_distance(t0, t1, _distance_upper_layer2(t0, t1))


def load_distance3_matrix(dictionary):
    global mat_distance3

    FILE = '/tmp/distance_3_m_%s.npy'%str(dictionary.version)

    if os.path.isfile(FILE):
        mat_distance3 = np.load(FILE)
    else:
        _compute_distance_3(dictionary)
        np.save(FILE, mat_distance3)


def distance3(t0, t1):
    if mat_distance3 is None:
        load_distance3_matrix(t0.dictionary)

    return mat_distance3[t0.index, t1.index]


def test_distance3(t):
    print("Distance from term %s -- %s"%(str(t), t.translations['fr']))

    other = sorted(((distance3(t, t1), t1) for t1 in Dictionary() if not t1.script.paradigm), reverse=True)
    kkk = [t for t in other]
    for d, tt in kkk[:30]:
        print("%s (%.2f) - %s [%s]"%(str(tt), d, tt.translations['fr'], ', '.join(t.relations.to(tt))))


def test_distance_same_layer(t):
    print("Distance from term %s -- %s"%(str(t), t.translations['fr']))

    other = sorted(((distance3(t, t1), t1) for t1 in Dictionary()), reverse=True)
    kkk = [t for t in other]
    for d, tt in kkk[:30]:
        print("%s (%.2f) - %s [%s]"%(str(tt), d, tt.translations['fr'], ', '.join(t.relations.to(tt))))


def ranking_from_term(term0, nb_terms=30):
    other = sorted([(distance3(term0, t1), t1) for t1 in term0.dictionary], reverse=True)
    return other[:nb_terms]


if __name__ == '__main__':
    # print(distance3(term(312), term(3112)))
    d = Dictionary(get_available_dictionary_version()[-1])
    # r = ranking_from_term(d.index[1])
    print('\n'.join("%s - %s (%f)"%(str(r[1]), r[1].translations['fr'], r[0]) for r in ranking_from_term(d.index[1])))





import inspect
import json
from collections import defaultdict
from itertools import product, groupby
import itertools
from bidict import bidict
import numpy as np
import os

from scipy.sparse.csr import csr_matrix
from ieml.ieml_objects.terms.relations import RELATIONS
from ieml.ieml_objects.terms.version import DictionaryVersion, get_default_dictionary_version

from ieml.ieml_objects.terms import Term
from ieml.script.constants import MAX_LAYER
from ieml.script.operator import script
from ieml.script.script import AdditiveScript, NullScript, MultiplicativeScript
from ieml.config import DICTIONARY_FOLDER

class InvalidDictionaryState(Exception):
    def __init__(self, dictionary, message):
        self.dictionary = dictionary
        self.message = message

    def __str__(self):
        return "Invalid state dictionary state for version %s: %s"%(str(self.dictionary.version), self.message)


class DictionarySingleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if len(args) < 1 or not isinstance(args[0], (DictionaryVersion, str)):
            version = get_default_dictionary_version()
        elif isinstance(args[0], DictionaryVersion):
            version = args[0]
        elif isinstance(args[0], str):
            version = DictionaryVersion(args[0])
        else:
            raise ValueError("Invalid argument for dictionary creation, expected dictionary version, not %s"%str(args[0]))

        if version not in cls._instances:
            cls._instances[version] = super(DictionarySingleton, cls).__call__(version, **kwargs)

        return cls._instances[version]


class Dictionary(metaclass=DictionarySingleton):
    def __init__(self, version, cache=True, load=True):
        super().__init__()

        if isinstance(version, str):
            version = DictionaryVersion(version)

        self.version = version

        self.cache = cache

        # static elements from the version object
        self.terms = None
        self.translations = None
        self.roots = None
        self.inhibitions = None

        self.index = None

        # computed elements stored in cache
        # terms -> int
        self.ranks = None
        # terms -> terms[] all the terms that decompose the key into subtable
        self.partitions = None
        self.parents = None
        # numpy array
        self.relations = None

        # list layer (int) -> list of terms at this layer
        self._layers = None         # make layers stacks
        self._singular_sequences = None
        self._singular_sequence_map = None

        if load:
            self.load()

    @property
    def is_build(self):
        return all(getattr(self, attr) is not None for attr in ['terms', 'translations', 'roots', 'inhibitions',
                                                                'index', 'ranks', 'partitions', 'parents', 'relations'])

    def load(self):
        cache_folder = os.path.join(DICTIONARY_FOLDER, 'cache')
        cache_json = os.path.join(cache_folder, 'cache_%s.json'%(str(self.version)))
        cache_relations = os.path.join(cache_folder, 'cache_%s_relations.npy'%(str(self.version)))

        if not os.path.isdir(cache_folder):
            os.makedirs(cache_folder)

        if not os.path.isfile(cache_json) or not os.path.isfile(cache_relations) or not self.cache:
            if not self.is_build:
                self.build()

            if self.cache:
                print("\t[*] Saving dictionary cache to disk (%s, %s)" % (str(cache_json), str(cache_relations)))
                state = self.__getstate__()
                # save relations as numpy array
                np.save(arr=state['relations'], file=cache_relations)
                del state['relations']

                with open(cache_json, 'w') as fp:
                    json.dump(state, fp)
        else:
            print("\t[*] Loading dictionary from disk (%s, %s)" % (str(cache_json), str(cache_relations)))

            with open(cache_json, 'r') as fp:
                state = json.load(fp)

            state['relations'] = np.load(cache_relations)
            self.__setstate__(state)

        print("\t[*] Dictionary loaded (version: %s, nb_roots: %d, nb_terms: %d)"%
              (str(self.version), len(self.roots), len(self)))

    def build(self):
        self._populate()
        self.compute_ranks()
        self.compute_relations()

    @property
    def singular_sequences(self):
        if self._singular_sequences is None:
            self._singular_sequences = sorted(self.terms[ss] for r in self.roots for ss in r.script.singular_sequences)

        return self._singular_sequences

    @property
    def layers(self):
        if self._layers is None:
            self._layers = [[] for _ in range(MAX_LAYER + 1)]
            for t in self.index:
                self._layers[t.script.layer].append(t)

        return self._layers

    @property
    def singular_sequences_map(self):
        if self._singular_sequence_map is None:
            self._singular_sequence_map = {ss: r for r in self.roots for ss in r.script.singular_sequences}

        return self._singular_sequence_map

    def compute_ranks(self):
        print("\t[*] Computing ranks")
        tables = defaultdict(list)
        self.ranks = {}

        def get_rank_partition(term0, term1):
            def is_connexe_tilling(coords, t):
                shape_t = t.cells.shape
                # test if the table table is a connexe tiling of the table t
                size = 1
                for i, indexes in enumerate(zip(*coords)):
                    # if i >= t.dim:
                    #     return False

                    indexes = sorted(set(indexes))
                    size *= len(indexes)

                    missing = [j for j in range(shape_t[i]) if j not in indexes]
                    # more than one transition from missing -> included
                    transitions = [j for j in missing if (j + 1) % shape_t[i] not in missing]
                    if len(transitions) > 1:
                        step = int(shape_t[i] / len(transitions))

                        # check if there is a periodicity
                        if any((j + step)%shape_t[i] not in transitions for j in transitions):
                            return False

                    # at least a square of 2x2x1
                    if len(indexes) < 2 and i != 2:
                        return False

                if len(coords) == size:
                    return True

                return False

            def is_dim_subset(coords, t):
                """
                Return if it is a dim subset (only one dimension, the others are not touched)
                If True, return (True, nb_dim)
                else (False, 0)

                :param coords:
                :param t:
                :return:
                """
                shape_ts0 = tuple(len(set(indexes)) for indexes in zip(*coords))
                shape_t = t.cells.shape

                if shape_ts0[2] != shape_t[2]:
                    # subset of tabs, return True if no subset in row/columns
                    return shape_ts0[0] == shape_t[0] and shape_ts0[1] == shape_t[1], shape_ts0[2]

                if shape_ts0[0] == shape_t[0]:
                    return True, shape_ts0[1]

                if shape_ts0[1] == shape_t[1]:
                    return True, shape_ts0[0]

                return False, 0

            if term0.script not in term1.script or self.ranks[term1] % 2 == 0:
                return None, None

            if len(term0.script.tables) != 1:
                # the paradigm is split between tables (weird case)
                # in this case, term0 can be a subset of tabs of each table of term1
                # its must match the tabs headers of term1 tables
                for t1 in term1.script.tables:
                    for t0 in term0.script.tables:
                        if len(t0.headers) != 1:
                            raise ValueError()

                        h = next(t0.headers.__iter__())

                        if h in t1.headers:
                            break
                    else:
                        break
                else:
                    # print("Tabs subset for %s in %s"%(str(term0), str(term1)))
                    return 0, None

            for table in tables[term1]:
                if term0.script not in table.paradigm:
                    continue

                coords = sorted(table.index(term0.script))

                is_dim, count = is_dim_subset(coords, table)
                if is_dim and (count == 1 or table.dim == 1):
                    # one dimension subset, it is a rank 3/5 paradigm
                    return 2, table

                # the coordinates sorted of the ss of s0 in the table t
                # rank is then 2/4
                if is_connexe_tilling(coords, table) or is_dim:
                    return 1, table
            return None, None

        def _define(term, rank, term_src, defined):
            self.ranks[term] = rank

            defined.add(term)
            self.partitions[term_src].add(term)
            self.parents[term] = term_src

            for t in term.script.tables:
                t_term = self.terms[t.paradigm]
                tables[t_term] += [t]

                self.ranks[t_term] = rank
                defined.add(t_term)
                self.partitions[term_src].add(t_term)
                self.parents[t_term] = term_src

                if len(t.headers) != 1:
                    for tab in t.headers:
                        self.ranks[self.terms[tab]] = rank
                        defined.add(self.terms[tab])
                        tables[self.terms[tab]] = self.terms[tab].tables
                        self.partitions[term_src].add(self.terms[tab])
                        self.parents[self.terms[tab]] = term_src

        self.partitions = defaultdict(set)
        self.parents = {}
        for ss in self.singular_sequences:
            self.ranks[ss] = 6

        for root in self.roots:
            defined = set()
            _define(root, 1, root, defined)

            # order by the cardinal (ieml order reversed)
            for term in sorted(self.roots[root], reverse=True)[1:]:
                if term in defined:
                    continue

                if not term.script.paradigm:
                    break

                res = defaultdict(list)
                for term1 in defined:
                    rank, parent_table = get_rank_partition(term, term1)
                    if rank is not None:
                        res[self.ranks[term1] + rank].append(term1)

                if not res:
                    raise ValueError("No rank candidates for %s" % str(term))
                if len(res) > 1:
                    # print("Multiple rank candidates for %s (%s)" %
                    #                  (str(term), ', '.join("%s:%d->%d"%(str(t1), self.ranks[t1], r) for r,v in res.items() for parent_t, t1 in v)))
                    print("Multiple rank candidates [%s] for the term %s. Choosing the maximum rank."%(", ".join(["%d/%s"%(r, str(t[0].script)) for r, t in res.items()]), str(term)))
                    m = max(res)
                    res = {m: res[m]}

                rank = next(res.__iter__())
                if len(res[rank]) > 1:
                    res[rank] = sorted(res[rank])
                    print("Multiple candidate for parent of %s with rank %d : [%s]. Choosing 1st one."%(str(term), rank, ", ".join(str(t) for t in res[rank])))
                term1 = res[rank][0]

                _define(term, rank, term1, defined)

            self.ranks[root] = 0

    def __len__(self):
        return len(self.terms)

    def __contains__(self, item):
        return script(item) in self.terms

    def __iter__(self):
        return self.index.__iter__()

    def get_root(self, script):
        try:
            res = {self.singular_sequences_map[ss] for ss in script.singular_sequences}
        except KeyError:
            return None

        if len(res) > 1:
            raise ValueError("Script %s is in multiples root paradigms [%s]" % (str(script), ', '.join(map(str, res))))

        return next(res.__iter__())

    def rel(self, type, term=None):
        if term:
            return [self.index[j] for j in self.relations[RELATIONS.index(type)][term.index, :].indices]
        else:
            return self.relations[RELATIONS.index(type)][:, :]

    def relations_graph(self, relations_types):
        if isinstance(relations_types, dict):
            res = np.zeros((len(self), len(self)), dtype=np.float)
            for reltype in relations_types:
                res += self.relations[RELATIONS.index(reltype)] * relations_types[reltype]
            return res

        return np.sum([self.relations[RELATIONS.index(reltype)] for reltype in relations_types])

    def compute_relations(self):
        print("\t[*] Computing relations")

        for i, t in enumerate(self.index):
            t.index = i

        _relations = {}
        contains = self._compute_contains()
        _relations['contains'] = contains
        _relations['contained'] = np.transpose(_relations['contains'])

        father = self._compute_father()

        for i, r in enumerate(['_substance', '_attribute', '_mode']):
            _relations['father' + r] = father[i, :, :]
            # _relations['child' + r] = children[i, :, :]

        siblings = self._compute_siblings()
        _relations['opposed'] = siblings[0]
        _relations['associated'] = siblings[1]
        _relations['crossed'] = siblings[2]
        _relations['twin'] = siblings[3]

        self._do_inhibitions(_relations)

        for i, r in enumerate(['_substance', '_attribute', '_mode']):
            _relations['child' + r] = np.transpose(_relations['father' + r])

        _relations['siblings'] = sum(siblings)
        _relations['inclusion'] = np.clip(_relations['contains'] + _relations['contained'], 0, 1)
        _relations['father'] = _relations['father_substance'] + _relations['father_attribute'] + _relations['father_mode']
        _relations['child'] = _relations['child_substance'] + _relations['child_attribute'] + _relations['child_mode']
        _relations['etymology'] = _relations['father'] + _relations['child']

        _relations['table'] = np.zeros((len(self), len(self)), dtype=np.float32)
        for i, t in enumerate(self._compute_table_rank()):
            _relations['table_%d'%i] = t
            _relations['table'] = np.maximum((i + 1.0) * t, _relations['table'])

        missing = {s for s in RELATIONS if s not in _relations}
        if missing:
            raise ValueError("Missing relations : {%s}"%", ".join(missing))

        self.relations = []
        for reltype in RELATIONS:
            self.relations.append(csr_matrix(_relations[reltype]))

    def _compute_table_rank(self):
        print("\t\t[*] Computing tables relations")

        _tables_rank = np.zeros((6, len(self), len(self)))

        for i, t in enumerate(self.index):
            if t.script.paradigm and len(t.script.tables) == 1 and t.script.tables[0].dim < 3:

                contained_ss = [self.terms[ss].index for ss in t.script.singular_sequences]
                index0, index1 = list(zip(*product(contained_ss, repeat=2)))
                _tables_rank[t.rank, index0, index1] = 1

                h = t.script.tables[0].headers[t.script]
                contained_r_c = [self.terms[s].index for s in itertools.chain(h.rows, h.columns) if s in self.terms]

                if contained_r_c:
                    index0, index1 = list(zip(*product(contained_r_c, repeat=2)))
                    _tables_rank[t.rank, index0, index1] = 1

        for root in self.roots:
            indexes = [p.index for p in self.roots[root]]
            index0, index1 = list(zip(*product(indexes, repeat=2)))
            _tables_rank[0, index0, index1] = 1

        return _tables_rank

    def _do_inhibitions(self, _relations):
        print("\t\t[*] Performing inhibitions")

        for r in self.roots:
            inhibitions = self.inhibitions[r]
            indexes = [t.index for t in self.roots[r]]
            # index0, index1 = list(zip(*product(indexes, repeat=2)))

            for rel in inhibitions:
                _relations[rel][indexes, :] = 0

    def _compute_contains(self):
        print("\t\t[*] Computing contains/contained relations")
        # contain/contained
        contains = np.diag(np.ones(len(self), dtype=np.int8))
        for r_p, v in self.roots.items():
            paradigms = {t for t in v if t.script.paradigm}

            for p in paradigms:
                _contains = [self.terms[ss].index for ss in p.script.singular_sequences] + \
                           [k.index for k in paradigms if k.script in p.script]
                contains[p.index, _contains] = 1

        return contains

    def _compute_father(self):
        print("\t\t[*] Computing father/child relations")

        def _recurse_script(script, res_indexes, depth):
            if isinstance(script, NullScript):
                return

            if script in self.terms:
                depth += 1
                res_indexes.append((self.terms[script].index, depth))

            if script.layer == 0:
                return

            for c in script.children:
                _recurse_script(c, res_indexes, depth)

        father = np.zeros((3, len(self), len(self)), dtype=np.float32)
        for t in self.terms.values():
            s = t.script

            for sub_s in s if isinstance(s, AdditiveScript) else [s]:
                if len(sub_s.children) == 0 or isinstance(sub_s, NullScript):
                    continue

                for i in range(3):
                    res_indexes = []
                    _recurse_script(sub_s.children[i], res_indexes, 0)
                    for j, d in res_indexes:
                        father[i, t.index, j] = 1.0/d**2

        return father

    def _compute_siblings(self):
        # siblings
        # 1 dim => the sibling type
        #  -0 opposed
        #  -1 associated
        #  -2 crossed
        #  -3 twin
        def _opposed_sibling(s0, s1):
            return not s0.empty and not s1.empty and\
                   s0.cardinal == s1.cardinal and\
                   s0.children[0] == s1.children[1] and s0.children[1] == s1.children[0]

        def _associated_sibling(s0, s1):
            return s0.cardinal == s1.cardinal and\
                   s0.children[0] == s1.children[0] and \
                   s0.children[1] == s1.children[1] and \
                   s0.children[2] != s1.children[2]

        def _crossed_sibling(s0, s1):
            return s0.layer >= 2 and \
                   s0.cardinal == s1.cardinal and \
                   _opposed_sibling(s0.children[0], s1.children[1]) and \
                   _opposed_sibling(s0.children[1], s1.children[0])

        siblings = np.zeros((4, len(self), len(self)), dtype=np.int8)

        print("\t\t[*] Computing siblings relations")

        for root in self.roots:
            if root.script.layer == 0:
                continue
            _twins = []

            for i, t0 in enumerate(self.roots[root]):
                if not isinstance(t0.script, MultiplicativeScript):
                    continue

                if t0.script.children[0] == t0.script.children[1]:
                    _twins.append(t0)

                for t1 in [t for j, t in enumerate(self.roots[root])
                           if j > i and isinstance(t.script, MultiplicativeScript)]:
                    if _opposed_sibling(t0.script, t1.script):
                        siblings[0, t0.index, t1.index] = 1
                        siblings[0, t1.index, t0.index] = 1

                    if _associated_sibling(t0.script, t1.script):
                        siblings[1, t0.index, t1.index] = 1
                        siblings[1, t1.index, t0.index] = 1

                    if _crossed_sibling(t0.script, t1.script):
                        siblings[2, t0.index, t1.index] = 1
                        siblings[2, t1.index, t0.index] = 1

            _twins = sorted(_twins, key=lambda t: t.script.cardinal)
            for card, g in groupby(_twins, key=lambda t: t.script.cardinal):
                twin_indexes = [t.index for t in g]

                if len(twin_indexes) > 1:
                    index0, index1 = list(zip(*product(twin_indexes, repeat=2)))
                    siblings[3, index0, index1] = 1

        return siblings

    def __getstate__(self):
        return {
            'relations': self.relations,
            'ranks': {str(t.script): r for t, r in self.ranks.items()},
            'partitions': {str(t.script): [str(tt.script) for tt in v] for t, v in self.partitions.items()},
            'parents': {str(t0.script): str(t1.script) for t0, t1 in self.parents.items()}
        }

    def __setstate__(self, state):
        self._populate(partitions=state['partitions'], ranks=state['ranks'], relations=state['relations'],
                       parents=state['parents'])

    def _populate(self, partitions=None, ranks=None, relations=None, parents=None):
        self.version.load()

        self.terms = {s: Term(s, dictionary=self) for s in self.version.terms}
        self.index = sorted(self.terms.values())

        for i, t in enumerate(self.index):
            t.index = i

        self.translations = {l: bidict({self.terms[t]: text for t, text in v.items()}) for l, v in
                             self.version.translations.items()}

        self.inhibitions = {self.terms[r]: v for r, v in self.version.inhibitions.items()}

        # if any(v for v in self.inhibitions.values()):
        #     raise InvalidDictionaryState("")

        self.roots = {self.terms[r]: [] for r in self.version.roots}

        try:
            for t in self.terms.values():
                self.roots[self.singular_sequences_map[t.script.singular_sequences[0]]].append(t)
        except KeyError as e:
            raise InvalidDictionaryState(self, "No root paradigm for script %s"%str(e.args[0]))

        if partitions is not None:
            self.partitions = defaultdict(set)
            for t, v in partitions.items():
                self.partitions[self.terms[t]] = {self.terms[tt] for tt in v}

        if ranks is not None:
            self.ranks = {}
            for t in self.index:
                self.ranks[t] = ranks[str(t.script)]

        if relations is not None:
            self.relations = relations

        if parents is not None:
            self.parents = {self.terms[t0]: self.terms[t1] for t0, t1 in parents.items()}

        # reset the cache properties
        self._layers = None
        self._singular_sequences = None

if __name__ == '__main__':
    # version = DictionaryVersion(datetime.date.today())
    # if False:
    #     d = Dictionary()
    #     version.terms = [str(t.script) for t in d.index]
    #     version.inhibitions = {str(t.script): v for t, v in d.inhibitions.items()}
    #     version.roots = [str(t.script) for t in d.roots]
    #     version.translations = {
    #         l: {
    #             str(t.script): trad for t, trad in v.items()
    #         } for l, v in d.translations.items()
    #     }
    #     version.upload_to_s3()

    Dictionary(DictionaryVersion()).build()

    # print(get_available_dictionary_version())

    # upload_dictionary_version(DictionaryVersion(datetime.date.today()))

    # get_dictionary_version("d")
    # Dictionary().build()
    # save_dictionary(DICTIONARY_FOLDER)





class InvalidIEMLObjectArgument(Exception):
    def __init__(self, type, msg):
        self.type = type
        self.message = msg

    def __str__(self):
        return 'Invalid arguments to create a %s object. %s'%(self.type.__name__, str(self.message))


class InvalidTreeStructure(Exception):
    def __init__(self, msg):
        self.message = msg

    def __str__(self):
        return 'Invalid tree structure. %s'%str(self.message)

class CantGenerateElement(Exception):
    def __init__(self, msg):
        self.message = msg

    def __str__(self):
        return 'Unable to generate element. %s'%self.message


from ieml.ieml_objects import Term, Word, Sentence, SuperSentence, Text, Hypertext
from ieml.script.tables import generate_tables
from ieml.script.operator import sc
from collections import namedtuple, defaultdict
from ieml.script.constants import AUXILIARY_CLASS, VERB_CLASS, NOUN_CLASS
from models.terms import TermsConnector
from models.relations.relations import RelationsConnector
import bidict
import numpy as np

ScoreNode = namedtuple('ScoreNode', ['script', 'score'])
Cache = namedtuple('Cache', ['source_layer', 'source_class', 'source_usl', 'usl_index'])
ParadigmMetadata = namedtuple('ParadigmMetadata', ['paradigm', 'score', 'nouns', 'auxiliary', 'verb'])


def _build_cache(usl_collection):

    def usl_elem_generator(u):
        if isinstance(u, (Word, Sentence, SuperSentence)):
            yield u
        elif isinstance(u, (Text, Hypertext)):
            for c in u.children:
                yield from usl_elem_generator(c)

    # For every term we associate a vector where each coordinate is the number of citations from a given layer
    # First coordinate: SuperSentence
    # Second coordinate: Sentence
    # Third coordinate: Word
    source_layer = defaultdict(lambda: [0, 0, 0])
    # For every term we associate a vector where each coordinate
    # is the number of citations from a given grammatical class
    source_class = defaultdict(lambda: [0, 0, 0])
    # Each term will be mapped to a vector indicating the number of citations from an usl in the usl_collection
    # each usl is given a coordinate in the usl_index bidict
    source_usl = defaultdict(lambda: [0 for u in usl_collection])
    # Every usl is given a coordinate for vector representation
    usl_index = bidict.bidict({i: u for i, u in enumerate(usl_collection)})

    for u in usl_collection:
        for elem in usl_elem_generator(u):
            terms = [term for term in elem.tree_iter() if isinstance(term, Term)]
            for t in terms:
                # TODO: Maybe use string representation of terms so that we don't need to worry about Terms/Script
                source_layer[t.script][coordinate[elem.__class__]] += 1
                source_class[t.script][elem.grammatical_class] += 1
                source_usl[t.script][usl_index.inv[u]] += layer_weight[elem.__class__]

                # We also count all the contained terms
                rc = RelationsConnector()
                term_rel = rc.get_script(t.script)
                if 'CONTAINS' in term_rel['RELATIONS']:
                    for child in term_rel['RELATIONS']['CONTAINS']:
                        child_script = sc(child)
                        source_layer[child_script][coordinate[elem.__class__]] += 1
                        source_class[child_script][elem.grammatical_class] += 1
                        source_usl[child_script][usl_index.inv[u]] += layer_weight[elem.__class__]


    return Cache(source_layer=source_layer, source_class=source_class, source_usl=source_usl, usl_index=usl_index.inv)


_cache = None


def rank_paradigms(paradigm_list, usl_collection):
    """
    Paradigms should be of type Script ***NOT TYPE Term***

    Parameters
    ----------
    paradigm_list
    usl_collection

    Returns
    -------

    """

    # TODO: So check and convert Term types to Script type
    result = []
    global _cache
    if not _cache:
        _cache = _build_cache(usl_collection)

    for paradigm in paradigm_list:

        score = (_cache.source_layer[paradigm][0] * layer_weight[SuperSentence] +
                 _cache.source_layer[paradigm][1] * layer_weight[Sentence] +
                 _cache.source_layer[paradigm][2] * layer_weight[Word])

        result.append(ParadigmMetadata(paradigm=paradigm, score=score,
                                       nouns=_cache.source_class[paradigm][NOUN_CLASS],
                                       auxiliary=_cache.source_class[paradigm][AUXILIARY_CLASS],
                                       verb=_cache.source_class[paradigm][VERB_CLASS]))

    return sorted(result, key=lambda x: x.score, reverse=True)


def rank_usls(term_list, usl_collection):
    """
    This method assumes the user will correctly giv it a list of root paradigms or of normal terms
    Parameters

    Paradigms should be of type Script ***NOT TYPE Term***
    ----------
    term_list: A list of terms, can be ROOT paradigms or plain old terms
    usl_collection

    Returns
    -------
    A dictionary that maps for each root paradigm a sorted list of USLs based on the number of citations
    """

    # TODO: So check and convert Term types to Script type

    global _cache
    if not _cache:
        _cache = _build_cache(usl_collection)

    return {term: sorted(usl_collection, key=lambda u: _cache.source_usl[term][_cache.usl_index[u]], reverse=True)
            for term in term_list}


def paradigm_usl_distribution(paradigm, usl_collection):
    """

    Parameters
    ----------
    paradigm
    usl_collection

    Returns
    -------
    Distribution of number of terms (contained in 'paradigm') cited overlaid on the paradigms table
    """

    global _cache
    if not _cache:
        _cache = _build_cache(usl_collection)

    if isinstance(paradigm, str):
        paradigm = sc(paradigm)

    tbls = generate_tables(paradigm)
    dist_tables = [np.zeros(table.cells.shape, dtype=int) for table in tbls]

    for tbl_idx, table in enumerate(tbls):
        it = np.nditer(table.cells, flags=['multi_index', 'refs_ok'])
        while not it.finished:
            # it[0] is the cell element
            dist_tables[tbl_idx][it.multi_index] += sum(_cache.source_layer[it[0].item()])
            it.iternext()

        # # We also want to count the cell citations coming from the headers
        # if table.dimension == 1:
        #     for i, row_header in enumerate(table.headers[0]):
        #         dist_tables[tbl_idx][i] += sum(_cache.source_layer[row_header])
        # elif table.dimension == 2:
        #     for i, row_header in enumerate(table.headers[0]):
        #         dist_tables[tbl_idx][i, :] += sum(_cache.source_layer[row_header])
        #     for i, col_header in enumerate(table.headers[1]):
        #         dist_tables[tbl_idx][:, i] += sum(_cache.source_layer[col_header])
        # elif table.dimension == 3:
        #     for i, row_header in enumerate(table.headers[0]):
        #         dist_tables[tbl_idx][i, :, :] += sum(_cache.source_layer[row_header])
        #     for i, col_header in enumerate(table.headers[1]):
        #         dist_tables[tbl_idx][:, i, :] += sum(_cache.source_layer[col_header])
        #     for i, tab_header in enumerate(table.headers[2]):
        #         dist_tables[tbl_idx][:, :, i] += sum(_cache.source_layer[tab_header])

    return dist_tables


coordinate = {
    SuperSentence: 0,
    Sentence: 1,
    Word: 2
}


layer_weight = {
    Word: 1,
    Sentence: 2,
    SuperSentence: 3
}

if __name__ == '__main__':

    paradigms_list = ["E:E:F:.", "E:F:.M:M:.-", "E:F:.O:O:.-"]

    #These 2 terms have the same root paradigm : E:E:F:.
    term_1 = Term(sc("E:E:F:."))
    term_2 = Term(sc("E:E:M:."))

    #The root paradigm of term_3 is E:F:.M:M:.-
    term_3 = Term(sc("E:M:.k.-"))

    usl_list1 = [term_1, term_2]
    usl_list2 = [term_3, term_1, term_3]
    usl_list3 = [term_1, term_3, term_2]

    term_1.check()
    term_2.check()
    term_3.check()

    paradigm_dico = rank_usls(paradigms_list,usl_list1)
    tc = TermsConnector()
    full_root_paradigms = tc.root_paradigms(ieml_only = True) # list of the 53 strings of the root paradigms

    paradigm_dico2 = rank_usls(paradigms_list, usl_list2)
    #self.assertTrue(len(paradigm_dico2) == len(full_root_paradigms))
    #self.assertTrue(len(paradigm_dico2["E:E:F:."]) == 1)
    #self.assertTrue(paradigm_dico2["E:F:.M:M:.-"] == 2)


from ieml.script.constants import MAX_SINGULAR_SEQUENCES


class InvalidScript(Exception):
    def __str__(self):
        return "Invalid arguments to create a script."


class InvalidScriptCharacter(InvalidScript):
    def __init__(self, character):
        self.character = character

    def __str__(self):
        return 'Invalid character %s for a parser.'%str(self.character)


class TooManySingularSequences(Exception):
    def __init__(self, num):
        self.num = num

    def __str__(self):
        return 'Too many singular sequences in the paradigms (%d, max %d).'%(self.num, MAX_SINGULAR_SEQUENCES)


class IncompatiblesScriptsLayers(InvalidScript):
    def __init__(self, s1, s2):
        self.s1 = s1
        self.s2 = s2

    def __str__(self):
        return 'Unable to add the two scripts %s, %s they have incompatible layers.'%(str(self.s1), str(self.s2))


class NoRemarkableSiblingForAdditiveScript(Exception):
    pass


from collections import defaultdict
from itertools import chain

import numpy

from ieml.ieml_objects.exceptions import InvalidIEMLObjectArgument
from ieml.ieml_objects.sentences import SuperSentence, Sentence, Clause, SuperClause
from ieml.ieml_objects.terms import Term
from ieml.ieml_objects.texts import Text
from ieml.ieml_objects.words import Word, Morpheme
from ieml.paths.exceptions import IEMLObjectResolutionError, ResolveError
from ieml.paths.parser.parser import PathParser
from ieml.paths.paths import Path, Coordinate, MultiplicativePath, AdditivePath, ContextPath


def path(p):
    if isinstance(p, Path):
        return p
    if isinstance(p, str):
        return PathParser().parse(p)

    try:
        l = list(p)
        if all(isinstance(e, Path) for e in l):
            res = l[0]
            for p in l[1:]:
                res *= p
            return res

    except TypeError:
        pass

    raise ValueError("Invalid argument to create a path.")


def _resolve_path_tree_graph(tree_graph, path):
    if isinstance(path, Coordinate):
        coords = [path]
    else:
        coords = list(path.children)

    result = set()
    c0 = coords[0]
    if c0.kind == 's':
        stack = {tree_graph.root}
    elif c0.kind == 'a':
        # tous les attributs
        stack = {attr[0] for vals in tree_graph.transitions.values() for attr in vals}
    else:
        # tous les modes
        return {attr[1][2] for vals in tree_graph.transitions.values() for attr in vals}

    for c in coords[1:]:
        _stack = set()
        for s in stack:
            if c.kind == 's':
                raise ValueError("Double substance s in path [%s]"%str(path))

            if c.kind == 'm':
                if c.index is not None:
                    _stack.add(tree_graph.transitions[s][c.index][1][2])
                else:
                    _stack |= {vals[1][2] for vals in tree_graph.transitions[s]}

            if c.kind == 'a':
                if c.index is not None:
                    _stack.add(tree_graph.transitions[s][c.index][0])
                else:
                    _stack |= {vals[0] for vals in tree_graph.transitions[s]}

        if not _stack:
            return set()

        stack = _stack

    return stack


def _resolve_path(obj, path):
    """path is a mul of coord or a coord"""
    if obj.__class__ not in path.context.accept:
        return set()

    if isinstance(obj, Text):
        if path.index:
            return {obj.children[path.index]}

        return set(obj.children)

    if isinstance(obj, (Sentence, SuperSentence)):
        return _resolve_path_tree_graph(obj.tree_graph, path)

    if isinstance(obj, Word):
        if path.kind == 'r':
            if path.index is not None:
                return {obj.root.children[path.index]}
            return set(obj.root.children)
        else:
            if path.index is not None:
                return {obj.flexing.children[path.index]}
            return set(obj.flexing.children)

    if isinstance(obj, Term):
        raise ValueError("Term not deferencable.")


def resolve(ieml_object, path):
    if ieml_object.__class__ not in path.context.accept:
        raise ValueError("Can't resolve [%s] on %s"%(str(path), str(ieml_object)))

    result = set()
    for d in path.develop:
        if isinstance(d, (Coordinate, MultiplicativePath)):
            result |= _resolve_path(ieml_object, d)
        else:
            # context path
            stack = {ieml_object}
            for c in d.children:
                _stack = set()
                for s in stack:
                    _stack |= _resolve_path(s, c)

                if not _stack:
                    break

                stack = _stack
            else:
                result |= stack

    return result


def _enumerate_paths(ieml_obj, level):
    if isinstance(ieml_obj, level):
        yield [], ieml_obj

    if isinstance(ieml_obj, Text):
        for i, t in enumerate(ieml_obj.children):
            for p, e in _enumerate_paths(t, level=level):
                yield [path('t%d'%i)] + p, e

    if isinstance(ieml_obj, (Sentence, SuperSentence)):
        for node in set(node for clause in ieml_obj for node in clause):
            for p, e in _enumerate_paths(node, level=level):
                yield [_tree_graph_path_of_node(ieml_obj.tree_graph, node)] + p, e

    if isinstance(ieml_obj, Word):
        for i, t in enumerate(ieml_obj.root.children):
            for p, e in _enumerate_paths(t, level=level):
                yield [path('r%d'%i)] + p, e

        if ieml_obj.flexing:
            for i, t in enumerate(ieml_obj.flexing.children):
                for p, e in _enumerate_paths(t, level=level):
                    yield [path('f%d' % i)] + p, e

    return


def _tree_graph_path_of_node(tree_graph, node):
    if node in tree_graph.nodes:
        nodes = [(node, False)]
    else:
        nodes = []

    # can be a mode
    nodes += [(c[0], True) for c_list in tree_graph.transitions.values() for c in c_list if c[1][2] == node]
    if not nodes:
        raise ValueError("Node not in tree graph : %s" % str(node))

    def _build_coord(node, mode=False):
        if node == tree_graph.root:
            return [Coordinate(kind='s')]

        parent = tree_graph.nodes[numpy.where(tree_graph.array[:, tree_graph.nodes_index[node]])[0][0]]

        return _build_coord(parent) + \
               [Coordinate(index=[c[0] for c in tree_graph.transitions[parent]].index(node), kind='m' if mode else 'a')]

    return AdditivePath([MultiplicativePath(_build_coord(node, mode)) for node, mode in nodes])


def enumerate_paths(ieml_obj, level=Term):
    for p, t in _enumerate_paths(ieml_obj, level=level):
        if len(p) == 1:
            yield p[0], t
        else:
            yield ContextPath(p), t


def _build_deps_tree_graph(rules):
    def _node():
        return {
            'resolve': defaultdict(_node),  # the indexed coordinate
            'context': [],
            'rules': defaultdict(_node)
        }

    # contain s, a et m (relative path to the root)
    roots = defaultdict(_node)

    s = Coordinate(kind='s')

    for p, e in rules:
        if isinstance(p, ContextPath):
            actual_P = p.children[0]  # multiplicative or coordinate
            ctx_P = path(p.children[1:])
        else:
            actual_P = p
            ctx_P = None

        if isinstance(actual_P, MultiplicativePath):
            actual_P = list(actual_P.children)
        else:
            actual_P = [actual_P]

        # actual_p is the list of coordinate to navigate
        # ctx_p is the rest of the path or None

        current_node = roots

        # replacing s0 -> s
        if actual_P[0].kind == 's' and actual_P[0].index == 0:
            actual_P[0] = s

        first = True
        for c in actual_P:
            if c.index is not None:
                categorie = 'resolve'
            else:
                categorie = 'rules'

            if first:
                current_node = current_node[c]
                first = False
            else:
                current_node = current_node[categorie][c]

        # the nodes with ctx_P is None will be instanciate
        current_node['context'].append((ctx_P, e))

    result = []

    def _merge_nodes(n0, n1):
        for cat in ('resolve', 'rules'):
            for r in n1[cat]:
                if r in n0[cat]:
                    _merge_nodes(n0[cat][r], n1[cat][r])
                else:
                    n0[cat][r] = n1[cat][r]

        # merge ctx
        n0['context'] += n1['context']

    coord_a = Coordinate('a')
    coord_m = Coordinate('m')

    # apply the rules on the tree starting from s
    def _apply_rule(node, _path, mode=False):
        # we add the globals rules to its own
        _merge_nodes(node['rules'][coord_a], roots[coord_a])
        _merge_nodes(node['rules'][coord_m], roots[coord_m])

        obj = _resolve_ctx(_path, node['context'])
        error = obj is None

        if mode or not node['resolve']:
            return error, obj

        # resolve the children
        max_i = max(n.index for n in node['resolve'])

        _result = {
            'a': [],
            'm': []
        }

        for r in ('a', 'm'):
            indexed = {k.index: node['resolve'][k] for k in node['resolve'] if k.kind == r}
            rules = [k for k in node['rules'] if k.kind == r]

            if max(indexed, default=-1) + 1 > len(indexed) + (1 if rules else 0):
                raise ResolveError("Not enough information to instantiate all the %s rules. At %s."%(r, _path))

            for i in range(max_i + 1):
                if i in indexed:
                    child = indexed[i]
                else:
                    child = _node()

                _merge_nodes(child, node['rules'][r])

                e, n = _apply_rule(child, _path + "%s%d"%(r, i), mode=r == 'm')
                error |= e
                _result[r].append(n)

        for a, m in zip(_result['a'], _result['m']):
            result.append((obj, a, m))

        return error, obj
    # 'context' attribute is now the sub element
    # generating the triplet from the root s

    error, o = _apply_rule(roots[s], 's0')

    return error, result[::-1]


def _build_deps_text(rules):
    indexes = defaultdict(list)
    generals = []

    for p, e in rules:
        if isinstance(p, ContextPath):
            actual_P = p.children[0]  # coordinate
            ctx_P = path(p.children[1:])
        else:
            actual_P = p
            ctx_P = None

        if not isinstance(actual_P, Coordinate):
            raise ResolveError("A text must be defined by a coordinate, not %s"%actual_P)

        if actual_P.index is not None:
            indexes[actual_P.index].append((ctx_P, e))
        else:
            generals.append((ctx_P, e))

    if not indexes:
        return False, []

    if max(indexes, default=-1) + 1 - len(indexes) > 1:
        # if there is more than one missing
        raise ResolveError("Index missing on text definition.")

    error = False
    i = 0
    result = []
    while i <= max(indexes):
        ctx_rules = generals[:]
        if i in indexes:
            ctx_rules.extend(indexes[i])

        node = _resolve_ctx("t%d"%i, ctx_rules)
        error |= node is None
        result.append(node)
        i += 1

    return error, result


def _build_deps_word(rules):
    result = {
        'r': {
            'indexes': defaultdict(list),
            'generals': []
        },
        'f': {
            'indexes': defaultdict(list),
            'generals': []
        }
    }

    for p, e in rules:
        if isinstance(p, ContextPath):
            # should be an error but support it
            actual_P = p.children[0]  # coordinate
            ctx_P = path(p.children[1:])
        else:
            actual_P = p
            ctx_P = None

        if not isinstance(actual_P, Coordinate):
            raise ResolveError("A word must be defined by a single coordinate.")

        if actual_P.index is not None:
            result[actual_P.kind]['indexes'][actual_P.index].append((ctx_P, e))
        else:
            result[actual_P.kind]['generals'].append((ctx_P, e))

    error = False
    for k in result:
        # k for kind (r or f)

        indexes = result[k]['indexes']
        generals = result[k]['generals']

        if not indexes and not generals:
            result[k] = []
            continue

        if max(indexes, default=-1) + 1 > len(indexes) + len(generals):
            # if there is more than one missing
            raise ResolveError("Index missing on word definition.")

        current = []
        length = len(indexes) + len(generals)
        generals = generals.__iter__()
        for i in range(length):
            if i in indexes:
                ctx_rules = indexes[i]
            else:
                ctx_rules = [next(generals)]

            node = _resolve_ctx("%s%d"%(k, i), ctx_rules)
            error |= node is None

            current.append(node)

        result[k] = current

    return error, result


def _inferred_types(path, e):
    result = set()
    for inf in path.context.switch:
        if e.__class__ in path.context.switch[inf]:
            result.add(inf)

    if result:
        return result

    raise ResolveError("No compatible type found with the path %s and the ieml object of type %s"%
                     (str(path), e.__class__.__name__))


_errors = []
_context_stack = []
debug = False


def _context_error_handler(func):
    def wrapper(_path, rules):
        _context_stack.append(_path)

        result = None
        if debug:
            result = func(rules)
        else:
            try:
                result = func(rules)
            except ResolveError as e:
                _errors.append((':'.join(_context_stack[1:]), str(e)))
            except InvalidIEMLObjectArgument as e:
                _errors.append((':'.join(_context_stack[1:]), str(e)))

        _context_stack.pop()

        return result

    return wrapper


@_context_error_handler
def _resolve_ctx(rules):
    """
    Resolve the context of the rules (the type of this element), and building the ieml element.
    :param rules:
    :return:
    """
    if not rules:
        raise ResolveError("Missing node definition.")

    # if rules == [(None, e)] --> e
    if len(rules) == 1 and rules[0][0] is None:
        return rules[0][1]

    if any(r[0] is None for r in rules):
        raise ResolveError("Multiple definition, multiple ieml object provided for the same node.")

    if any(not isinstance(r[0], Path) for r in rules):
        raise ResolveError("Must have only path instance.")

    # resolve all the possible types for this element
    r0 = rules[0]
    types = _inferred_types(*r0)
    for r in rules[1:]:
        types = types.intersection(_inferred_types(*r))

    if not types:
        raise ResolveError("No definition, no type inferred on rules list.")

    if len(types) > 1:
        raise ResolveError("Multiple definition, multiple type inferred on rules list.")

    type = next(types.__iter__())

    if type == Word:
        error, deps = _build_deps_word(rules)
        if error:
            return

        flexing = None
        if deps['f']:
            flexing = Morpheme(deps['f'])
        if not deps['r']:
            raise ResolveError("No root for the word node.")

        return Word(Morpheme(deps['r']), flexing)

    if type == Text:
        error, deps = _build_deps_text(rules)
        if error:
            return

        return Text(deps)

    if type in (SuperSentence, Sentence):
        error, deps = _build_deps_tree_graph(rules)
        if error:
            return

        if type == Sentence:
            clauses = []
            for s, a, m in deps:
                clauses.append(Clause(s, a, m))
            return Sentence(clauses)
        else:
            clauses = []
            for s, a, m in deps:
                clauses.append(SuperClause(s, a, m))
            return SuperSentence(clauses)

    raise ResolveError("Invalid type inferred %s"%type.__name__)


def resolve_ieml_object(paths, elements=None):
    global _errors, _context_stack
    _errors = []
    _context_stack = []

    if isinstance(paths, dict):
        paths = list(paths.items())

    if elements is None:
        result = _resolve_ctx('', [(d, e) for p, e in paths for d in p.develop])
    else:
        result = _resolve_ctx('', [(d, e) for e, p in zip(elements, paths) for d in p.develop])

    if _errors:
        raise IEMLObjectResolutionError(_errors)

    return result




from _operator import mul
from collections import namedtuple, defaultdict
from itertools import groupby, chain, product

from functools import reduce

from ieml.ieml_objects.sentences import SuperSentence, Sentence
from ieml.ieml_objects.terms import Term
from ieml.ieml_objects.texts import Text
from ieml.ieml_objects.words import Word
from ieml.paths.constants import COORDINATES_KINDS
from ieml.paths.exceptions import PathError

Context = namedtuple("Context", ['accept', 'conserve', 'switch'])

COORDINATES_CONTEXTS = {
    't': Context(accept={Text}, conserve=False, switch={
        Text: {SuperSentence, Sentence, Word, Term}
    }),
    'a': Context(accept={SuperSentence, Sentence}, conserve=True, switch={
        SuperSentence: {Sentence},
        Sentence: {Word}
    }),
    's': Context(accept={SuperSentence, Sentence}, conserve=True, switch={
        SuperSentence: {Sentence},
        Sentence: {Word}
    }),
    'm': Context(accept={SuperSentence, Sentence}, conserve=False, switch={
        SuperSentence: {Sentence},
        Sentence: {Word}
    }),
    'r': Context(accept={Word}, conserve=False, switch={
        Word: {Term}
    }),
    'f': Context(accept={Word}, conserve=False, switch={
        Word: {Term}
    })
}


class Path:
    # represent a path.
    # A path is an element of the form
    # this correspond to a descent in an ieml tree
    #
    # 't0:sa0:r0'
    # 't:s:(r0+f1)'
    # 't:(sm0+sa0:m0):r1'
    # '(t0+t1):s:s:f'
    #
    # A path have a context, the types of ieml-object that this path can deference.
    # The type of the ieml_object inferred by this path on a well typed ieml-object is determined by the context at the
    # end of the string.
    #
    # There are 5 different context, for each ieml-object type. A context switch occur with ':'.
    # - the Text context :
    #    't' is the only literal. If indexed like 'tn' where 'n' is an integer, it represent the n-th element in the
    # text object. This context allow only one character ('t' or 'tn') then it is always followed by ':' or nothing.
    # It can be followed by a super-sentence, a sentence or a word context.
    #
    # - the Super Sentence and Sentence ctx:
    # Begin with 's' or 'sn' but if the 'n' is an integer, it will be ignored. Can be followed by 'an' or 'mn'
    # where 'n' is an integer. When it is followed by 'an' it lead to the the n-th attribute of the current node. Same
    # for 'mn', but for the mode of the n-th attribute (the clause of this node, the n-th attribute and his mode). If
    # it is a mode, it can only be followed by a context switch ':'.
    # Supersentence lead to Sentence ctx and sentence to word ctx.
    #
    # - the Word context:
    # Only one literal of the form 'r' | 'rn' | 'f' | 'fn' where 'n' is an integer. These literal cannot be followed by
    # another character. If 'r' it lead all the root terms (the morphem), 'rn' lead only the n-th in the root morphem.
    # Same for 'f'|'fn' and the flexing morphem.
    #
    #  Ex: 't0:sa0:r' -> In a text context (or usl), the first element of the text, which is a sentence,
    #  then in a sentence context, the word of the first attribute of the root (index on 's' is ignored),
    #  then in a word context, the roots terms.
    #
    #
    #
    # the context of a path can be accessed:
    #   p.context.accept is a set of accepted types to deference
    #   p.context.switch is a dict of accepted types to inferred types i.e. the type of b = a[p] where a type is in
    # p.context.accept. len(p.context.switch) == len(p.context.accept)
    #
    # We add the '+' notation. An additive path is defined by the sum of multiple paths.
    # To get the list of sub path : p.develop
    #
    # p.cardinal == len(p.develop)
    #
    #


    def __init__(self, children):
        try:
            children = list(children)
        except TypeError:
            raise ValueError("The children must be iterable.")

        if any(not isinstance(child, Path) for child in children):
            raise ValueError('The children must be path instances.')

        # children unfold
        _children = []
        for c in children:
            if isinstance(c, self.__class__):
                _children += c.children
            else:
                if len(c.children) == 1:
                    _children.append(c.children[0])
                else:
                    _children.append(c)

        self.children = tuple(_children)
        self.cardinal = None
        self._development = None
        self.context = None

    @property
    def develop(self):
        if self._development is None:
            if self.cardinal == 1:
                self._development = (self,)
            else:
                self._development = self._develop()

        return self._development

    def _develop(self):
        raise NotImplemented

    def _resolve_context(self):
        if isinstance(self, Coordinate):
            self.context = COORDINATES_CONTEXTS[self.kind]
            return

        # the development is either a mul of coords or a context of mul and coords

        accept = set()
        switch = defaultdict(set)
        conserve = False

        for d in self.develop:
            if isinstance(d, Coordinate):
                accept |= d.context.accept
                conserve = conserve or d.context.conserve
                for c in d.context.accept:
                    switch[c] |= d.context.switch[c]

            elif isinstance(d, MultiplicativePath):
                # must ensure that the context is conserved until the last element
                for ctx in d.children[0].context.accept:
                    for c in d.children[:-1]:
                        # if the coord not conserve the current context
                        if not c.context.conserve:
                            raise PathError("Invalid developed path, context discontinuity.", d)

                        if ctx not in c.context.accept:
                            # the context is lost
                            break
                    else:
                        last = d.children[-1].context
                        # check the last coord if it accept the element
                        if ctx in last.accept:
                            # if the last element conserve the context, set it to true (at least one path conserve)
                            conserve = conserve or last.conserve

                            # add the passed type to the accept list
                            accept |= {ctx}

                            # add the switch of the last element
                            switch[ctx] |= last.switch[ctx]

            else:
                # instance of context path
                for str_acc in d.children[0].context.accept:
                    stack = {str_acc}
                    for c in d.children:
                        _stack = set()
                        for s in stack:
                            if s in c.context.accept:
                                _stack |= c.context.switch[s]

                        if not _stack:
                            break

                        stack = _stack
                    else:
                        accept |= {str_acc}
                        switch[str_acc] |= stack

        if not accept:
            raise PathError("No context match this path.", self)

        self.context = Context(accept=accept, conserve=conserve, switch=dict(switch))

    def __mul__(self, other):
        if isinstance(other, str):
            from .tools import path
            other = path(other)

        if isinstance(other, Path):
            return ContextPath([self, other])

        raise NotImplemented

    def __add__(self, other):
        if isinstance(other, str):
            from .tools import path
            other = path(other)

        if isinstance(other, Path):
            return AdditivePath([self, other])

        raise NotImplemented

    def __eq__(self, other):
        return str(self) == str(other)

    def __hash__(self):
        return hash(self.__str__())


class ContextPath(Path):
    def __init__(self, children):
        if not children:
            raise ValueError('Must be a non empty children.')

        super().__init__(children)

        if len(self.children) < 2:
            raise ValueError("A context path must have at least two children.")

        self.cardinal = reduce(mul, [c.cardinal for c in self.children])
        self._resolve_context()

    def __str__(self):
        result = []
        for c in self.children:
            if isinstance(c, AdditivePath):
                result.append('(%s)'%str(c))
            else:
                result.append(str(c))

        return ':'.join(result)

    def _develop(self):
        return tuple(ContextPath(p) for p in product(*[c.develop for c in self.children]))


class AdditivePath(Path):
    def __init__(self, children):
        if not children:
            raise ValueError('Must be a non empty children.')

        super().__init__(children)

        self.cardinal = sum(c.cardinal for c in self.children)
        self._resolve_context()

    def __str__(self):
        return '+'.join(map(str, self.children))

    def _develop(self):
        return tuple(chain.from_iterable(c.develop for c in self.children))


class MultiplicativePath(Path):
    def __init__(self, children):
        if not children:
            raise ValueError('Must be a non empty children.')

        super().__init__(children)

        self.cardinal = reduce(mul, [c.cardinal for c in self.children])

        self._resolve_context()

    def __str__(self):
        result = ''
        for c in self.children:
            if isinstance(c, AdditivePath):
                result += '(%s)'%str(c)
            else:
                result += str(c)

        return result

    def _develop(self):
        return tuple(MultiplicativePath(p) for p in product(*[c.develop for c in self.children]))


class Coordinate(Path):
    def __init__(self, kind, index=None):

        if not isinstance(kind, str) or not len(kind) == 1 or not kind in COORDINATES_KINDS:
            raise ValueError("A coordinate kind must be one of the following (%s)."%
                             ', '.join(map(str, COORDINATES_KINDS)))

        self.kind = kind

        if index is not None:
            if not isinstance(index, int):
                raise ValueError("Coordinate index must be int.")

            self.index = index
        else:
            self.index = None

        # no children
        super().__init__(())

        self.cardinal = 1

        self._resolve_context()

    def __str__(self):
        return self.kind + (str(self.index) if self.index is not None else '')

